# -*- coding: utf-8 -*-
"""
    Message bus - placeholder code until proper message broker is in place.
"""
from __future__ import absolute_import

import logging


log = logging.getLogger(__name__)


class MessageBus(object):

    def __init__(self, app=None):
        self.app = None
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if not hasattr(app, 'extensions'):
            app.extensions = {}

        app.extensions['messagebus'] = self
        self._consumers = {}

    def register_consumer(self, callback, queue_name):
        self._consumers.setdefault(queue_name, []).append(callback)

    def publish_message(self, queue_name, message):
        for consumer in self._consumers.get(queue_name, []):
            consumer(queue_name, message)


def drift_init_extension(app, **kwargs):
    app.messagebus = MessageBus(app)



# -*- coding: utf-8 -*-
"""
Schames listing APIs
"""
import logging

from flask import current_app, url_for
from flask.views import MethodView
from flask_smorest import Blueprint, abort
import marshmallow as ma


log = logging.getLogger(__name__)


bp = Blueprint('schemas', 'Schemas', url_prefix='/schemas', description='Something about schemas')


def drift_init_extension(app, api, **kwargs):
    api.register_blueprint(bp)


@bp.route('/', endpoint='schemas')
class SchemaListAPI(MethodView):

    """Fabular"""

    def get(self):
        """
        List schemas

        Returns a list of schema media type names and an `href` to the schema
        object.
        """
        json_schema = current_app.extensions.get("jsonschema", None)
        if not json_schema:
            return []  # Return code 200, 0 hits.

        rs = [
            {
                "media_type_name": media_type_name,
                "href": url_for(
                    "schema.schema",
                    media_type_name=media_type_name,
                    _external=True
                ),
            }
            for media_type_name in json_schema.get_media_type_names()
        ]

        return rs


@bp.route('/<string:media_type_name>', endpoint='schema')
class SchemaAPI(MethodView):

    def get(self, media_type_name):
        """
        Find a single schema

        Returns the JSON scema object for the given media type.
        """
        json_schema = current_app.extensions.get("jsonschema", None)
        if json_schema:
            schema_object = json_schema.get_schema_for_media_type(
                media_type_name)
            return schema_object
        else:
            return abort(404)



# -*- coding: utf-8 -*-
"""
This module contains generic and application-level sql logic.
"""
import datetime
import logging

from sqlalchemy import text
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy import Column, DateTime

log = logging.getLogger(__name__)

utc_now = text("(now() at time zone 'utc')")

Base = declarative_base()


# ! TODO: Move contents to resources.postgres
# These are here due to importing from other modules
from drift.core.resources.postgres import get_sqlalchemy_session, sqlalchemy_session  # noqa: F401


class ModelBase(Base):
    __abstract__ = True

    # we use this declared_attr method to ensure that these automatic columns are placed at the end
    @declared_attr
    def create_date(cls):
        return Column(DateTime, nullable=False, server_default=utc_now, index=False)

    @declared_attr
    def modify_date(cls):
        return Column(DateTime, nullable=False, server_default=utc_now,
                      onupdate=datetime.datetime.utcnow)

    def as_dict(self):
        """
        Returns the data for the row as a dictionary
        """
        columns = self.__table__.columns._data.keys()
        data = {}
        for c in columns:
            data[c] = getattr(self, c)
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module converts an AWS API Gateway proxied request to a WSGI request.
Pretty much copied verbatim from https://github.com/logandk/serverless-wsgi
"""
import base64
import os
import sys
import logging
from werkzeug.datastructures import Headers
from werkzeug.wrappers import Response
from werkzeug.urls import url_encode
from werkzeug._compat import BytesIO, string_types, to_bytes, wsgi_encoding_dance

# List of MIME types that should not be base64 encoded. MIME types within `text/*`
# are included by default.
TEXT_MIME_TYPES = [
    "application/json",
    "application/javascript",
    "application/xml",
    "application/vnd.api+json",
]


# The logger is already configured at this point by the lambda thunker so we need to reset it.
root = logging.getLogger()
if root.handlers:
    for handler in root.handlers:
        root.removeHandler(handler)

logging.basicConfig(format='[%(levelname)s] %(name)s: %(message)s', level=logging.INFO)


# Log out import and app factory exceptions explicitly with traceback because the AWS
# host is not terribly keen on doing so.
try:
    from drift.flaskfactory import drift_app
    app = drift_app()
except Exception:
    logging.exception("Can't create Drift app object.")


def handler(event, context):
    return handle_request(app, event, context)


def all_casings(input_string):
    """
    Permute all casings of a given string.
    A pretty algoritm, via @Amber
    http://stackoverflow.com/questions/6792803/finding-all-possible-case-permutations-in-python
    """
    if not input_string:
        yield ""
    else:
        first = input_string[:1]
        if first.lower() == first.upper():
            for sub_casing in all_casings(input_string[1:]):
                yield first + sub_casing
        else:
            for sub_casing in all_casings(input_string[1:]):
                yield first.lower() + sub_casing
                yield first.upper() + sub_casing


def handle_request(app, event, context):

    # This document contains info on the format and possible values of 'event'.
    # https://docs.aws.amazon.com/apigateway/latest/developerguide/set-up-lambda-proxy-integrations.html

    if event.get("source") in ["aws.events", "serverless-plugin-warmup"]:
        return {}

    headers = Headers(event[u"headers"])

    if u"amazonaws.com" in headers.get(u"Host", u""):
        script_name = "/{}".format(event[u"requestContext"].get(u"stage", ""))
    else:
        script_name = ""

    # If a user is using a custom domain on API Gateway, they may have a base
    # path in their URL. This allows us to strip it out via an optional
    # environment variable.
    path_info = event[u"path"]
    base_path = os.environ.get("API_GATEWAY_BASE_PATH", "")
    if base_path:
        script_name = "/" + base_path

        if path_info.startswith(script_name):
            path_info = path_info[len(script_name) :]

    body = event[u"body"] or ""
    if event.get("isBase64Encoded", False):
        body = base64.b64decode(body)
    if isinstance(body, string_types):
        body = to_bytes(body, charset="utf-8")

    environ = {
        "API_GATEWAY_AUTHORIZER": event[u"requestContext"].get(u"authorizer"),
        "CONTENT_LENGTH": str(len(body)),
        "CONTENT_TYPE": headers.get(u"Content-Type", ""),
        "PATH_INFO": path_info,
        "QUERY_STRING": url_encode(event.get(u"queryStringParameters") or {}),
        "REMOTE_ADDR": event[u"requestContext"]
        .get(u"identity", {})
        .get(u"sourceIp", ""),
        "REMOTE_USER": event[u"requestContext"]
        .get(u"authorizer", {})
        .get(u"principalId", ""),
        "REQUEST_METHOD": event[u"httpMethod"],
        "SCRIPT_NAME": script_name,
        "SERVER_NAME": headers.get(u"Host", "lambda"),
        "SERVER_PORT": headers.get(u"X-Forwarded-Port", "80"),
        "SERVER_PROTOCOL": "HTTP/1.1",
        "event": event,
        "context": context,
        "wsgi.errors": sys.stderr,
        "wsgi.input": BytesIO(body),
        "wsgi.multiprocess": False,
        "wsgi.multithread": False,
        "wsgi.run_once": False,
        "wsgi.url_scheme": headers.get(u"X-Forwarded-Proto", "http"),
        "wsgi.version": (1, 0),
    }

    # AWS API Gateway overwrite the X-Forwarded-For header which means that we need to pass
    # the real remote IP from nginx using X-Real-IP header. Just to be on the safe side we
    # make sure that it's really nginx calling the API Gateway:
    # TODO: Make a better check for this:
    apirouter_calling_us = environ['REMOTE_ADDR'].startswith('10.')

    if not environ['REMOTE_ADDR'] or apirouter_calling_us:
        environ['REMOTE_ADDR'] = headers.get(u"X-Real-IP", u"")
        headers['X-Forwarded-For'] = environ['REMOTE_ADDR']

    for key, value in environ.items():
        if isinstance(value, string_types):
            environ[key] = wsgi_encoding_dance(value)

    for key, value in headers.items():
        key = "HTTP_" + key.upper().replace("-", "_")
        if key not in ("HTTP_CONTENT_TYPE", "HTTP_CONTENT_LENGTH"):
            environ[key] = value

    response = Response.from_app(app, environ)

    # If there are multiple Set-Cookie headers, create case-mutated variations
    # in order to pass them through APIGW. This is a hack that's currently
    # needed. See: https://github.com/logandk/serverless-wsgi/issues/11
    # Source: https://github.com/Miserlou/Zappa/blob/master/zappa/middleware.py
    new_headers = [x for x in response.headers if x[0] != "Set-Cookie"]
    cookie_headers = [x for x in response.headers if x[0] == "Set-Cookie"]
    if len(cookie_headers) > 1:
        for header, new_name in zip(cookie_headers, all_casings("Set-Cookie")):
            new_headers.append((new_name, header[1]))
    elif len(cookie_headers) == 1:
        new_headers.extend(cookie_headers)

    returndict = {u"statusCode": response.status_code, u"headers": dict(new_headers)}

    if response.data:
        mimetype = response.mimetype or "text/plain"
        if (
            mimetype.startswith("text/") or mimetype in TEXT_MIME_TYPES
        ) and not response.headers.get("Content-Encoding", ""):
            returndict["body"] = response.get_data(as_text=True)
        else:
            returndict["body"] = base64.b64encode(response.data).decode("utf-8")
            returndict["isBase64Encoded"] = "true"

    return returndict




# -*- coding: utf-8 -*-
"""
    drift - Tenancy logic
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Figure out which tenant is in play
"""
from __future__ import absolute_import

import os
import logging
import socket
import warnings

from flask import request, has_request_context
from werkzeug.local import LocalProxy


log = logging.getLogger(__name__)


def drift_init_extension(app, **kwargs):
    pass


def get_tenant_name(request_headers=None):
    """Figure out tenant name in the following order:
    1. Tenant name explicitly specified in request header under the key 'Drift-Header'.
    2. Tenant name is the left most part of the host name specified in the request header.
    3. Default tenant name is specified in environment variable 'DRIFT_DEFAULT_TENANT'.
    """
    if request_headers:
        if 'Drift-Tenant' in request_headers:
            return request_headers['Drift-tenant']
        if 'Host' in request_headers:
            host_parts = split_host(request_headers['Host'])
            if host_parts['tenant']:
                return host_parts['tenant']

    return os.environ.get('DRIFT_DEFAULT_TENANT')


# Flask specific block begins
def _flask_get_tenant_name():
    """Calls get_tenant_name with request header and host if applicable.
    """
    request_headers = request.headers if has_request_context() else None
    return get_tenant_name(request_headers)


current_tenant_name = LocalProxy(_flask_get_tenant_name)


# Deprecated name as it's a bit misleading.
def _get_tenant_from_hostname(*args, **kw):
    # warnings.warn(
    #     "'tenant_from_hostname' has been renamed 'current_tenant_name'.",
    #     DeprecationWarning,
    #     stacklevel=2
    # )
    return _flask_get_tenant_name()

tenant_from_hostname = LocalProxy(_get_tenant_from_hostname)
# Flask specific block ends


def split_host(host):
    """Split 'host' into tenant, domain and port if possible."""
    ret = {
        'host': host,
        'tenant': None,
        'domain': None,
        'port': None,
    }

    # IPv6 barf because of https://tools.ietf.org/html/rfc3986#section-3.2.2
    if '[' in host and ']' in host:
        return ret  # Quickly and dirtily assume IPv6

    if ':' in host:
        ret['host'], ret['port'] = host.split(':', 1)
    if '.' in ret['host'] and not _is_valid_ipv4_address(ret['host']):
        ret['tenant'], ret['domain'] = host.split('.', 1)

    return ret


def _is_valid_ipv4_address(address):
    try:
        socket.inet_pton(socket.AF_INET, address)
    except AttributeError:  # no inet_pton here, sorry
        try:
            socket.inet_aton(address)
        except socket.error:
            return False
        return address.count('.') == 3
    except socket.error:  # not a valid address
        return False

    return True




import os
import sys
import logging
import importlib
import json
import os.path
import time
import warnings
import pkgutil

from flask import Flask
from flask_smorest import Api
from werkzeug.middleware.proxy_fix import ProxyFix
from werkzeug.utils import import_string, ImportStringError
from drift.fixers import ReverseProxied, CustomJSONEncoder
from drift.utils import get_app_root
import drift.core.extensions

log = logging.getLogger(__name__)


APP_INIT_THRESHOLD = .250  # If app creation takes longer a warning is logged.
MODULE_INIT_THRESHOLD = .050  # If module initialization takes longer a warning is logged.


class AppRootNotFound(RuntimeError):
    """To enable CLI to filter on this particular case."""
    pass


def drift_app(app=None):
    try:
        return _drift_app(app=app)
    except Exception:
        log.exception("Flask app creation failed.")


def _drift_app(app=None):
    """Flask factory for Drift based apps."""

    app_root = get_app_root()
    app = app or Flask('drift', instance_path=app_root, root_path=app_root)
    app.url_map.strict_slashes = False

    log.info("Init app.instance_path: %s", app.instance_path)
    log.info("Init app.static_folder: %s", app.static_folder)
    log.info("Init app.template_folder: %s", app.template_folder)

    # load deployment settings
    flask_config = load_flask_config()
    for k in flask_config.keys():
        if k in os.environ:
            flask_config[k] = os.environ[k]
    app.config.update(flask_config)

    try:
        cfg = import_string('config.config')
        app.config.from_object(cfg)
        log.info("Overriding configuration in config/config.json with config/config.py")
    except ImportStringError:
        log.info("No config.py in config folder. This is fine.")

    _apply_patches(app)

    # Install apps, api's and extensions.
    sys.path.insert(0, app_root)  # Make current app available
    app.config['OPENAPI_VERSION'] = "3.0.2"
    api = Api(app)

    # shitmixing this since flask-rest-api steals the 301-redirect exception
    def err(*args, **kwargs):
        pass

    api._register_error_handlers = err

    t = time.time()

    with app.app_context():
        install_modules(app, api)

    elapsed = time.time() - t
    if elapsed > APP_INIT_THRESHOLD:
        log.warning("Module installation took %.3f seconds.", elapsed)

    return app


_sticky_app_config = None


def set_sticky_app_config(app_config):
    """Assign permanently 'app_config' as the one and only app config. Useful for tests."""
    global _sticky_app_config
    _sticky_app_config = app_config


def load_flask_config(app_root=None):
    if _sticky_app_config is not None:
        return _sticky_app_config

    app_root = app_root or get_app_root()
    config_filename = os.path.join(app_root, 'config', 'config.json')
    if not os.path.exists(config_filename):
        raise AppRootNotFound("No config file found at: '{}'".format(config_filename))

    log.info("Loading configuration from %s", config_filename)
    with open(config_filename) as f:
        config_values = json.load(f)

    config_values['app_root'] = app_root
    return config_values


def enumerate_plugins(config):
    """Returns a list of resource, extension and app module names for current deployable."""

    # Include explicitly referenced resource modules.
    resources = config.get("resources", [])

    # Include all core extensions and those referenced in the config.
    pkgpath = os.path.dirname(drift.core.extensions.__file__)
    extensions = [
        'drift.core.extensions.' + name
        for _, name, _ in pkgutil.iter_modules([pkgpath])
        if name not in ['test', 'tests'] and not name.startswith('test_')
    ]
    extensions += config.get("extensions", [])
    extensions = sorted(list(set(extensions)))  # Remove duplicates

    # Include explicitly referenced app modules
    apps = config.get("apps", [])

    return {
        'resources': resources,
        'extensions': extensions,
        'apps': apps,
        'all': resources + extensions + apps,
    }


def install_modules(app, api):
    """Install built-in and product specific apps and extensions."""

    plugins = enumerate_plugins(app.config)
    resources, extensions, apps = plugins['resources'], plugins['extensions'], plugins['apps']

    log.info(
        "Installing extras:\nResources:\n\t%s\nExtensions:\n\t%s\nApps:\n\t%s",
        "\n\t".join(resources), "\n\t".join(extensions), "\n\t".join(apps)
    )
    # print(
    #     "Installing extras:\nResources:\n\t%s\nExtensions:\n\t%s\nApps:\n\t%s" %
    #     ("\n\t".join(resources), "\n\t".join(extensions), "\n\t".join(apps))
    # )

    performance = []
    # first, try new-style install of the plugins
    # The order of initialization matters, so we try both new and old style here.
    resources = init_plugin_list(app, api, resources, 'resources', performance)
    extensions = init_plugin_list(app, api, extensions, 'extensions', performance)
    apps = init_plugin_list(app, api, apps, 'apps', performance)

    # then, continue with old-style init of different kinds of apps
    # backwards compatibility
    for module_name in apps:
        init_legacy_module(app, module_name, performance)

    # Report import time if they are terrible

    report = [
        "\t{total_time:.3f} ({import_time:.3f})\t{module_name} [{category}]".format(**module)
        for module in performance
        if module['total_time'] > MODULE_INIT_THRESHOLD
    ]

    if report:
        log.warning("Performance issues:\n%s" % '\n'.join(report))


def init_legacy_module(app, module_name, performance):
    t = time.time()
    m = importlib.import_module(module_name)
    blueprint_name = "{}.blueprints".format(module_name)

    try:
        m = importlib.import_module(blueprint_name)
    except ImportError:
        log.exception("Can't import blueprint %s", blueprint_name)
    else:
        # This is a deprecated import mechanism
        warnings.warn(
            "extensions should initialize using 'drift_inít_extension()', "
            "not using a 'blueprints.py' module import.",
            DeprecationWarning)
        import_time = time.time() - t
        try:
            m.register_blueprints(app)
        except Exception:
            log.exception("Couldn't register blueprints for module '%s'", module_name)

        performance.append({
            'category': 'legacy',
            'module_name': module_name,
            'total_time': time.time() - t,
            'import_time': import_time,
        })


def init_plugin_list(app, api, plugin_names, category, performance):
    """
    Walk through a list of plugins and initialize them new-style,
    returning a list of those plugins that weren't initialized
    """
    result = []
    for plugin in plugin_names:
        success = False
        try:
            success = init_single_plugin(app, api, plugin, category, performance)
        except Exception as e:
            log.exception("Exception in init_single_plugin for %s: %s", plugin, e)
            raise
        if not success:
            result.append(plugin)
    return result


def init_single_plugin(app, api, module_name, category, performance):
    """
    Import a single plugin and call its initialization function
    """
    init = False
    t = time.time()
    m = importlib.import_module(module_name)
    import_time = time.time() - t

    if hasattr(m, "drift_init_extension"):
        # the following takes a single 'app' argument and then a kwargs dict
        m.drift_init_extension(app, api=api)
        init = True
    elif hasattr(m, "register_extension"):
        m.register_extension(app)
        init = True
    else:
        if category in ['apps', 'extensions']:
            log.warning(
                "%s module '%s' has no drift_init_extension() function.",
                category.title(), module_name
            )

    performance.append({
        'category': category,
        'module_name': module_name,
        'total_time': time.time() - t,
        'import_time': import_time,
    })

    return init


def _apply_patches(app):
    """
    Apply special fixes and/or monkey patches. These fixes will hopefully
    become obsolete with proper fixes made to these libraries in the future.
    """
    # "X-Forwarded-For" remote_ip fixup when running behind a load balancer.
    # Normal setup on AWS includes a single ELB which appends to "X-Forwarded-For"
    # header value, thus one proxy.
    num_proxies = 1
    app.wsgi_app = ProxyFix(app.wsgi_app, x_for=num_proxies, x_proto=num_proxies, x_host=num_proxies)

    # Fixing SCRIPT_NAME/url_scheme when behind reverse proxy (i.e. the
    # API router).
    app.wsgi_app = ReverseProxied(app.wsgi_app)

    # Datetime fix
    app.json_encoder = CustomJSONEncoder

    # Make all json pretty
    app.config['JSONIFY_PRETTYPRINT_REGULAR'] = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2016, Florent Thiery
# Run uwsgi with '--profiler pycall' and run the parser on uwsgi.log

import sys
from collections import Counter

from click import echo

HEADER = '##############'
PATTERN = "[uWSGI Python profiler"
TOP = 20


def main():

    with open(sys.argv[1]) as f:
        echo('Reading file')
        d = f.read().strip()

    calls = list()
    modules = list()
    profiler_name = 'pyline'

    echo('Splitting lines')
    lines = d.split('\n')
    echo('Parsing lines')
    for l in lines:
        if l.startswith(PATTERN):
            if ' CALL: ' in l:
                # pycall
                profiler_name = 'pycall'
                module = l.split(' CALL: ')[1].split(' (')[0]
                call = l.split(' ')[-5]
            else:
                # pyline
                module = l.split(' file ')[1].split(' line ')[0]
                call = l.split(' ')[-2]
            calls.append("%s > %s" % (module, call))
            modules.append(module)

    def count(items, items_name='calls'):
        if len(items) != 0:
            echo('Counting %s' % items_name)
            counter = Counter(items)
            echo('\n%s\nTop %s %s\n%s\n' % (HEADER, TOP, items_name, HEADER))
            items_count = len(items)
            top = counter.most_common(TOP)
            for t in top:
                percent = int(round(100*t[1]/items_count))
                if percent > 0:
                    echo("{} {}%".format(t[0], percent))

    echo('uWSGI python profiler %s results' % profiler_name)
    count(calls, 'calls')
    count(modules, 'modules')


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import sys
import os
import os.path
import subprocess

import boto3
from click import echo
from six import print_
from six.moves import input

from drift.utils import get_config


def get_options(parser):
    parser.add_argument("service", help="Service or deployable to connect to", nargs='?')


def run_command(args):
    service = args.service
    conf = get_config()
    tier_name = conf.tier['tier_name']
    region = conf.tier['aws']['region']
    ssh_key_name = conf.tier['aws']['ssh_key']
    deployables = conf.table_store.get_table('deployables').find({"tier_name": tier_name})
    deployables = {depl["deployable_name"]: depl for depl in deployables}
    if service is None:
        echo("Select an instance to connect to:")
        for k in sorted(deployables.keys()):
            echo("    {}".format(k))
        return

    elif service not in deployables:
        echo("Warning! Service or deployable '{}' not one of {}.".format(service, ", ".join(deployables.keys())))

    ssh_key_file = os.path.expanduser('~/.ssh/{}.pem'.format(ssh_key_name))

    # Get IP address of any instance of this deployable.
    sess = boto3.session.Session(region_name=region)
    ec2 = sess.client("ec2")
    filters = [{"Name": "instance-state-name", "Values": ["running"]},
               {"Name": "tag:tier", "Values": [tier_name]},
               {"Name": "tag:service-name", "Values": [service]},
               ]
    echo("Getting a list of EC2's from AWS matching the following criteria:")
    for criteria in filters:
        echo("   {} = {}".format(criteria["Name"], criteria["Values"][0]))

    ret = ec2.describe_instances(Filters=filters)
    instances = []
    for res in ret["Reservations"]:
        instances += res["Instances"]

    if not instances:
        echo("No instance found which matches the criteria.")
        return

    echo("Instances:")
    inst = instances[0]
    for i, ins in enumerate(instances):
        lb = [tag["Value"] for tag in ins["Tags"] if tag["Key"] == "launched-by"] or ["n/a"]
        echo("  {}: {} at {} launched by {} on {}".format(
            i + 1, ins["InstanceId"], ins["PrivateIpAddress"], lb[0], ins["LaunchTime"]))

    if len(instances) > 1:
        which = input("Select an instance to connect to (or press enter for first one): ")
        if which:
            inst = instances[int(which) - 1]
    else:
        echo("Only one instance available. Connecting to it immediately..")

    ip_address = inst["PrivateIpAddress"]
    cd_cmd = ""
    if service in deployables:
        cd_cmd = 'cd /etc/opt/{}; exec bash --login'.format(service)
    cmd = ["ssh", "ubuntu@{}".format(ip_address), "-i", ssh_key_file, "-t", cd_cmd]
    echo("\nSSH command: " + " ".join(cmd))
    p = subprocess.Popen(cmd)
    stdout, _ = p.communicate()
    if p.returncode != 0:
        if stdout:
            print_(stdout.decode())
        sys.exit(p.returncode)



# -*- coding: utf-8 -*-
"""
Schames listing APIs
"""
import logging
import importlib
from six.moves.http_client import SERVICE_UNAVAILABLE

from flask import current_app, g
from flask.views import MethodView
from flask_smorest import Blueprint, abort
import marshmallow as ma


log = logging.getLogger(__name__)
bp = Blueprint('healtchcheck', 'HealthCheck', url_prefix='/healthcheck', description='Service and tenant health check')


class HealthCheckSchema(ma.Schema):
    result = ma.fields.Str(description="Is the service healthy")


def drift_init_extension(app, api, **kwargs):
    api.register_blueprint(bp)


@bp.route('', endpoint='health')
class HealthCheckAPI(MethodView):
    no_jwt_check = ["GET"]

    @bp.response(HealthCheckSchema)
    def get(self):
        """
        Tenant health check

        Returns 200 if the service and tenant are in a good state or 503 if there is a problem.
        The caller does not need to be authenticated to call this endpoint.
        """
        # If there is no tenant, this health check is only reporting a successful rest call
        if not g.conf.tenant:
            return {'result': "ok, but no tenant specified."}

        if g.conf.tenant["state"] != "active":
            abort(SERVICE_UNAVAILABLE, ("Tenant is in state '%s' but needs to be 'active'." % g.conf.tenant["state"]))

        resources = current_app.config.get("resources")
        if not resources:
            abort(SERVICE_UNAVAILABLE, "Deployable is missing 'resources' section in drift config.")

        for module_name in resources:
            m = importlib.import_module(module_name)
            if hasattr(m, "healthcheck"):
                m.healthcheck()

        return {'result': "all is fine"}



#!/usr/bin/env python
import sys
import os
import argparse
import importlib
import getpass
from datetime import datetime
import logging
import subprocess
import socket
import boto3

from click import echo

from drift.management.gittools import get_branch, get_commit, get_repo_url, get_git_version
from drift.utils import pretty, set_pretty_settings, PRETTY_FORMATTER, PRETTY_STYLE
from driftconfig.util import get_default_drift_config_and_source, ConfigNotFound
from drift.flaskfactory import AppRootNotFound


def get_commands():
    commands = [
        f[:-3]
        for f in os.listdir(os.path.join(os.path.dirname(__file__), "commands"))
        if not f.startswith("_") and f.endswith(".py")
    ]
    return commands


def execute_cmd():
    try:
        return do_execute_cmd(sys.argv[1:])
    except AppRootNotFound as e:
        # A very common case that needs pretty printing
        echo(str(e))
    except KeyboardInterrupt:
        echo(" Aborting because you said so.")


def do_execute_cmd(argv):
    valid_commands = get_commands()
    parser = argparse.ArgumentParser(description="")

    parser.add_argument(
        '--tier',
        help="Specify which tenant to use. Will override any other settings."
    )
    parser.add_argument(
        '--tenant', '-t',
        help="Specify which tenant to use. Will override any other settings."
    )
    parser.add_argument(
        '--config',
        help="Specify which config source to use. Will override 'DRIFT_CONFIG_URL' environment variable."
    )
    parser.add_argument(
        "--loglevel", '-l',
        help="Logging level name. Default is WARNING.", default='WARNING'
    )
    parser.add_argument(
        '--formatter',
        help="Specify which formatter to use for text output. Default is {}.".format(
            PRETTY_FORMATTER)
    )
    parser.add_argument(
        '--style',
        help="Specify which style to use for text output. Default is {}.".format(
            PRETTY_STYLE)
    )

    parser.add_argument("-v", "--verbose", help="I am verbose!", action="store_true")
    subparsers = parser.add_subparsers(help="sub-command help", dest="cmd")
    subparsers.required = True
    for cmd in valid_commands:
        module = importlib.import_module("drift.management.commands." + cmd)
        subparser = subparsers.add_parser(cmd, help="Subcommands for {}".format(cmd))
        if hasattr(module, "get_options"):
            module.get_options(subparser)
        subparser.set_defaults(func=module.run_command)

    args = parser.parse_args(argv)

    if args.loglevel:
        logging.basicConfig(level=args.loglevel)

    if args.config:
        os.environ['DRIFT_CONFIG_URL'] = args.config

    try:
        conf, source = get_default_drift_config_and_source()
        echo("Drift configuration source: {!r}".format(source))
    except ConfigNotFound:
        pass

    set_pretty_settings(formatter=args.formatter, style=args.style)

    if args.tier:
        os.environ['DRIFT_TIER'] = args.tier
        echo("Tier set to {!r}.".format(args.tier))

    if args.tenant:
        os.environ['DRIFT_DEFAULT_TENANT'] = args.tenant
        echo("Default tenant set to {!r}.".format(args.tenant))

    if 'DRIFT_APP_ROOT' in os.environ:
        echo("App root set: DRIFT_APP_ROOT={!r}".format(os.environ['DRIFT_APP_ROOT']))

    args.func(args)


def get_app_version():
    """
    Return the version of the current app.
    It's gotten by running: python setup.py --version
    """

    # HACK: Get app root:
    from drift.utils import get_app_root
    app_root = get_app_root()

    p = subprocess.Popen(
        [sys.executable, 'setup.py', '--version'],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        cwd=app_root
    )
    out, err = p.communicate()
    out, err = (str(s.decode("utf-8")) for s in (out, err))
    if p.returncode != 0:
        raise RuntimeError(
            "Can't get version of this deployable. Error: {} - {}".format(p.returncode, err)
        )

    version = out.strip()
    return version


def check_connectivity(instances):
    SSH_PORT = 22
    for inst in instances:
        ip_address = inst.private_ip_address
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(2)
        result = sock.connect_ex((ip_address, SSH_PORT))
        if result != 0:
            raise RuntimeError("Unable to connect to '%s'. Is your VPN connection active?" % ip_address)


def get_ec2_instances(region, tier, service_name):
    """
    Returns all EC2 instances on the specified region, tier and service.
    Raises an error if any of the instances are not reachable in SSH
    """
    filters = {
        'tag:service-name': service_name,
        "instance-state-name": "running",
        "tag:tier": tier,
    }
    echo("Finding ec2 instances in region {!r} from filters: {!r}".format(region, filters))

    conn = boto3.client('ec2', region_name=region)

    reservations = conn.get_all_reservations(filters=filters)
    instances = [i for r in reservations for i in r.instances]

    if not instances:
        raise RuntimeError("Found no running ec2 instances in region '%s', tier '%s' and service '%s'" % (region, tier, service_name))

    check_connectivity(instances)

    return instances


def create_deployment_manifest(method, comment=None, deployable_name=None):
    """Returns a dict describing the current deployable."""

    git_version = get_git_version()
    git_commit = get_commit()

    info = {
        'method': method,
        'deployable': deployable_name,
        'version': get_app_version(),
        'username': getpass.getuser(),
        'comment': comment,
        'datetime': datetime.utcnow().isoformat(),

        'git_branch': get_branch(),
        'git_commit': git_commit,
        'git_commit_url': get_repo_url() + "/commit/" + git_commit,
        'git_release': git_version['tag'] if git_version else 'untagged-branch',
    }

    return info



# -*- coding: utf-8 -*-
import unittest

from driftconfig.util import get_drift_config
from drift.core.extensions.apikeyrules import get_api_key_rule
from drift.tests import DriftTestCase
from drift.systesthelper import setup_tenant


class ApiKeyRulesTest(DriftTestCase):

    @classmethod
    def setUpClass(cls):

        config_size = {
            'num_org': 5,
            'num_tiers': 2,
            'num_deployables': 4,
            'num_products': 2,
            'num_tenants': 2,
        }

        ts = setup_tenant(config_size=config_size, use_app_config=False)
        cls.ts = ts
        cls.tier_name = ts.get_table('tiers').find()[0]['tier_name']
        cls.product_name = ts.get_table('products').find()[0]['product_name']
        cls.tenant_name_1 = ts.get_table('tenant-names').find({'product_name': cls.product_name})[0]['tenant_name']
        cls.tenant_name_2 = ts.get_table('tenant-names').find({'product_name': cls.product_name})[1]['tenant_name']
        cls.deployable_1 = ts.get_table('deployables').find({'tier_name': cls.tier_name})[0]['deployable_name']
        cls.conf = get_drift_config(
            ts=ts,
            tenant_name=cls.tenant_name_1,
            tier_name=cls.tier_name,
            deployable_name=cls.deployable_1,
        )
        cls._add_rules()

    @classmethod
    def _add_rules(cls):
        """The location of this function is for convenience as the test function comes right after it."""

        # Make a few rules to test various cases:
        # Rule 1: reject clients 1.6.0 and 1.6.2 and ask them to upgrade.
        # Rule 2: redirect client 1.6.5 to another tenant.
        # Rule 3: always let client 1.6.5, and 1.6.6 pass through.
        # Rule 4: reject all clients with message "server is down".
        rules = [
            ('upgrade-client-1.6', ["1.6.0", "1.6.2"], 'reject', [404, {"action": "upgrade_client"}]),
            ('redirect-to-new-tenant', ["1.6.5"], 'redirect', cls.tenant_name_2),
            ('always-pass', ["1.6.5", "1.6.6"], 'pass', cls.tenant_name_1),
            ('downtime-message', [], 'reject', [503, {"message": "The server is down for maintenance."}]),
        ]

        api_key_rules = cls.ts.get_table('api-key-rules')

        for i, (rule_name, version_pattern, rule_type, custom) in enumerate(rules):
            row = api_key_rules.add({
                'product_name': cls.product_name,
                'rule_name': rule_name,
                'assignment_order': i,
                'version_patterns': version_pattern,
                'rule_type': rule_type,
            })
            if rule_type == 'reject':
                row['reject'] = {'status_code': custom[0], 'response_body': custom[1]}
            elif rule_type == 'redirect':
                row['redirect'] = {'tenant_name': custom}

            row['response_header'] = {'Test-Rule-Name': rule_name}  # To test response headers

        cls.ts.get_table('api-keys').add({
            'api_key_name': cls.product_name,
            'product_name': cls.product_name
        })

        cls.ts.get_table('api-keys').add({
            'api_key_name': 'custom-key',
            'key_type': 'custom',
        })

    def test_custom_key(self):
        # Only keys of type 'product' need to be associated with a product. Custom keys can
        # be universal and thus implicitly excempt from any product rule.
        headers = {
            "Drift-Api-Key": "custom-key",
        }
        url = "https://%s.kaleo.io/drift" % self.tenant_name_1
        rule = get_api_key_rule(headers, url, self.conf)
        self.assertIsNone(rule)

    def test_no_api_key_means_pass(self):
        # The reason for this is that some endpoints don't need
        # a key and this is enforced by nginx before it gets here
        url = "https://%s.kaleo.io/drift" % self.tenant_name_1
        rule = get_api_key_rule({}, url, self.conf)
        self.assertIsNone(rule)

    def test_pass_rule_passes(self):
        headers = {
            "Drift-Api-Key": "%s:%s" % (self.product_name, "1.6.6")
        }
        url = "https://%s.kaleo.io/drift" % self.tenant_name_1
        rule = get_api_key_rule(headers, url, self.conf)
        self.assertIsNotNone(rule)
        self.assertIsNone(rule["status_code"])

    def test_reject_rule_rejects(self):
        headers = {
            "Drift-Api-Key": "%s:%s" % (self.product_name, "1.6.1")
        }
        url = "https://%s.kaleo.io/drift" % self.tenant_name_1
        rule = get_api_key_rule(headers, url, self.conf)
        self.assertIsNotNone(rule)
        self.assertEqual(rule['status_code'], 503)
        self.assertEqual(rule['response_body']['message'], "The server is down for maintenance.")

    def test_upgrade_rule_passes_with_info(self):
        headers = {
            "Drift-Api-Key": "%s:%s" % (self.product_name, "1.6.0")
        }
        url = "https://%s.kaleo.io/drift" % self.tenant_name_1
        rule = get_api_key_rule(headers, url, self.conf)
        self.assertIsNotNone(rule)
        self.assertEqual(rule['status_code'], 404)
        self.assertEqual(rule['response_body']['action'], "upgrade_client")

    def test_redirect_rule_redirects(self):
        headers = {
            "Drift-Api-Key": "%s:%s" % (self.product_name, "1.6.5")
        }
        url = "https://%s.kaleo.io/drift" % self.tenant_name_1
        rule = get_api_key_rule(headers, url, self.conf)
        self.assertIsNotNone(rule)
        self.assertEqual(rule['status_code'], 307)
        self.assertEqual(rule['response_header']['Location'], "https://%s.kaleo.io/drift" % self.tenant_name_2)

    def test_redirect_rule_passes_when_redirected(self):
        headers = {
            "Drift-Api-Key": "%s:%s" % (self.product_name, "1.6.5")
        }
        url = "https://%s.kaleo.io/drift" % self.tenant_name_2
        rule = get_api_key_rule(headers, url, self.conf)
        self.assertIsNotNone(rule)
        self.assertIsNone(rule['status_code'])


if __name__ == '__main__':
    unittest.main()



import sys
import os.path

from click import echo
from fabric import Connection, Config
from drift.management import get_ec2_instances
from drift.utils import get_config

EC2_USERNAME = 'ubuntu'
APP_LOCATION = r"/usr/local/bin/{}"
SERVICE_PORT = 10080


def get_options(parser):
    parser.add_argument("cmd", help="Command to run on all instances", nargs='?', default=None)
    parser.add_argument(
        "--ip",
        help="Deploy to a certain instance instead of across the cluster")
    parser.add_argument(
        "--upload",
        help="Upload a file to an instance."
    )


def run_command(args):
    cmd = args.cmd
    if not cmd and args.upload is None:
        echo("Please enter command to run. Example: kitrun.py remotecmd \"ls -l\"")
        return

    conf = get_config()
    if not conf.deployable:
        echo("Deployable '{}' not found in config '{}'.".format(
            conf.drift_app['name'], conf.domain['domain_name']))
        sys.exit(1)

    service_name = conf.deployable['deployable_name']
    tier = conf.tier['tier_name']
    region = conf.tier['aws']['region']
    ssh_key_name = conf.tier['aws']['ssh_key']
    ssh_key_file = os.path.expanduser('~/.ssh/{}.pem'.format(ssh_key_name))

    echo("\n*** EXECUTING REMOTE COMMAND '{}' ON SERVICE '{}' / TIER '{}' IN REGION '{}'\n".format(cmd, service_name, tier, region))

    instances = get_ec2_instances(region, tier, service_name)

    for ec2 in instances:
        ip_address = ec2.private_ip_address

        echo("*** Running '{}' on {}...".format(cmd, ip_address))

        conf = Config()
        conf.connect_kwargs.key_filename = ssh_key_file
        conn = Connection(host=ip_address, user=EC2_USERNAME, config=conf)
        if args.upload:
            conn.put(args.upload, "~/")
        else:
            conn.run(cmd)
        echo()



"""
    drift - tracking setup code
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Setup tracking metadata for the http requests and responses
"""
from __future__ import absolute_import

import logging
import json
import uuid
import traceback
import sys

from six import StringIO
from flask import make_response, jsonify, request, current_app
from werkzeug.exceptions import HTTPException
from flask_smorest import abort

from drift.core.extensions.jwt import query_current_user, jwt_not_required

log = logging.getLogger(__name__)


def drift_init_extension(app, **kwargs):

    # app.handle_exception = partial(handle_all_exceptions, app.handle_exception)
    # app.handle_user_exception = partial(handle_all_exceptions, app.handle_user_exception)

    @app.errorhandler(500)
    @app.errorhandler(503)
    def deal_with_exceptions(e):
        return handle_all_exceptions(e)

    @app.errorhandler(400)
    @app.errorhandler(401)
    @app.errorhandler(403)
    @app.errorhandler(404)
    @app.errorhandler(405)
    @app.errorhandler(422)
    def deal_with_aborts(e):
        return handle_all_exceptions(e)

    # Health check endpoint. Always succeedes.
    @jwt_not_required
    @app.route('/healthcheck', methods=['GET'])
    def gurko_handler():
        return jsonify({"status": "fine"})

    # Borko endpoint. Always fails.
    @jwt_not_required
    @app.route('/borko', methods=['GET', 'POST'])
    def borko_handler():
        data = request.get_json()
        if data:
            status_code = data.pop('status_code', 400)
            abort(int(status_code), **data)
        else:
            raise RuntimeError("Borko raising an error.")


def handle_all_exceptions(e):
    is_server_error = not isinstance(e, HTTPException)

    ret = {}
    error = {}
    ret['error'] = error

    if is_server_error or e.code >= 500:
        # Use context_id from the client if it's available, or make one if not.
        log_context = request.headers.get("Drift-Log-Context")
        log_context = json.loads(log_context) if log_context else {}
        context_id = log_context.get("request_id", str(uuid.uuid4()).replace("-", ""))
        error['context_id'] = context_id
        title = str(e) + " - [{}]".format(context_id)

    if is_server_error:
        # Do a traceback if caller has dev role, or we are running in debug mode.
        current_user = query_current_user()
        if (current_user and "dev" in current_user['roles']) or current_app.debug:
            sio = StringIO()
            ei = sys.exc_info()
            sio.write("%s: %s\n" % (type(e).__name__, e))
            traceback.print_exception(ei[0], ei[1], ei[2], None, sio)
            error["traceback"] = sio.getvalue()
            sio.close()
            error['description'] = str(e)
        else:
            error['description'] = "Internal Server Error"

        # The exception is logged out and picked up by Splunk or comparable tool.
        # The 'context_id' in the title enables quick cross referencing with the
        # response body below.
        log.exception(title)

        ret['status_code'] = 500
        ret['message'] = "Internal Server Error"
        error['code'] = 'internal_server_error'
    else:
        ret['status_code'] = e.code
        ret['message'] = e.name
        error['code'] = 'user_error' if e.code < 500 else 'server_error'
        error['description'] = e.description

        # Support for Flask Restful 'data' property in exceptions.
        if hasattr(e, 'data') and e.data:
            error.update(e.data)
            if 'schema' in error:
                del error['schema']

            # Legacy field 'message'. If it's in the 'data' payload, rename the field
            # to 'description'.
            if 'message' in e.data:
                error['description'] = error.pop('message')

            # Legacy field ´code´ has been renamed error_code. the "code" keyword is used
            # in flask_smorest.abort to signify status code.
            if 'error_code' in e.data:
                error['code'] = error.pop('error_code')

        if e.code >= 500:
            # It's a "soft" server error. Let's log it out.
            log.warning(title + " " + error['description'])
    return make_response(jsonify(ret), ret['status_code'])



#!/usr/bin/env python
import sys
import os
import os.path
import subprocess
import time
import shutil
import yaml
import json

import boto3
import click
from click import echo, secho
from jinja2 import Environment, FileSystemLoader
from driftconfig.config import get_redis_cache_backend
from drift.utils import get_config, pretty


@click.command()
@click.option(
    '--tier-name', '-t', help='Tier name.'
)
@click.option('--offline', '-o', is_flag=True, help="Run serverless-offline.")
@click.option('--preview', '-p', is_flag=True, help="Preview, do not run 'sls' command.")
@click.option('--verbose', '-v', is_flag=True, help="Verbose output.")
@click.option('--clean', '-c', is_flag=True, help="Clean autogenerated files and folders and exit.")
@click.version_option('1.0')
def cli(tier_name, offline, preview, verbose, clean):
    """
    Generate settings for Serverless lambdas and deploy to AWS.
    """
    now = time.time()
    filename_yaml = 'serverless.yml'

    if clean:
        os.unlink(filename_yaml)
        if os.path.exists('./package-lock.json'):
            os.unlink('./package-lock.json')
        shutil.rmtree('.serverless', ignore_errors=True)
        shutil.rmtree('./node_modules', ignore_errors=True)
        secho("Cleaned up all autogenerated files and folders.", fg='green')
        sys.exit(0)

    secho("Deploy serverless Drift app", bold=True)

    # Make sure .gitignore is ignoring the autogenerated fluff
    with open(".gitignore") as f:
        gitignore = f.read()

    if '# Serverless' not in gitignore:
        gitignore = "# Serverless\n"\
            "serverless.yml\n"\
            ".serverless/\n"\
            "node_modules/\n"\
            "package.json\n"\
            "package-lock.json\n"\
            ".env\n\n" + gitignore
        with open(".gitignore", "w") as f:
            f.write(gitignore)

    conf = get_config(tier_name=tier_name)
    ts = conf.table_store
    tier = conf.tier
    tier_name = tier['tier_name']
    secho("Processing {}".format(tier_name))

    if 'aws' not in tier:
        secho("Error: 'aws' entry missing in tiers config. Exiting.", fg='red')
        sys.exit(1)

    for required_in_aws in ['region', 's3_bucket']:
        if required_in_aws not in tier['aws']:
            secho("Error: 'aws.{}' entry missing in tiers config. Exiting.".format(required_in_aws, fg='red'))
            sys.exit(1)

    # Figure out in which aws region this config is located
    aws_region = tier['aws']['region']
    ec2 = boto3.resource('ec2', aws_region)
    filters = [
        {'Name': 'tag:tier', 'Values': [tier_name]},
        {
            'Name': 'tag:Name',
            'Values': [
                tier_name + '-private-subnet-1',
                tier_name + '-private-subnet-2',
            ],
        },
    ]
    subnets = list(ec2.subnets.filter(Filters=filters))
    vpc_id = subnets[0].vpc_id
    subnet_ids = [subnet.id for subnet in subnets]

    filters = [
        {'Name': 'tag:tier', 'Values': [tier_name]},
        {'Name': 'tag:Name', 'Values': [tier_name + '-private-sg']},
    ]

    security_groups = list(ec2.security_groups.filter(Filters=filters))
    security_groups = [sg.id for sg in security_groups]

    # To auto-generate Redis cache url, we create the Redis backend using our config,
    # and then ask for a url representation of it:
    config_url = get_redis_cache_backend(ts, tier_name).get_url()

    # Find Cloudwatch log forwarding lambda
    tags_client = boto3.client('resourcegroupstaggingapi', region_name=aws_region)
    functions = tags_client.get_resources(
        TagFilters=[
            {
                'Key': 'tier', 'Values': ['DEVNORTH']
            },
            {
                'Key': 'service-type', 'Values': ['log-forwarder']
            },
        ],
        ResourceTypeFilters=['lambda:function']
    )['ResourceTagMappingList']
    log_forwarding_arn = functions[0]['ResourceARN'] if functions else None

    # Sum it up
    #
    # Template input parameters:
    # tier:
    #     tier_name
    #     config_url
    #     aws_region
    #     vpc_id
    #     security_groups
    #     subnets
    # deployable:
    #     deployable_name
    #
    # apihandler: api gateway handler or None
    #
    # events: (array of:)
    #     function_name     Actual Python function name, must be unique for the deployable
    #     event_type        One of s3, schedule, sns, sqs
    #
    #     # S3 specifics
    #     bucket            bucket name
    #
    #     # schedule specifics
    #     rate              rate or cron https://amzn.to/2yFynEA
    #
    #     # sns specifics
    #     topicName
    #
    #     # sqs specifics
    #     arn                arn:aws:sqs:region:XXXXXX:myQueue
    #     batchSize          10

    template_args = {
        'tier': {
            'tier_name': tier_name,
            'config_url': config_url,
            'aws_region': aws_region,
            'vpc_id': vpc_id,
            'security_groups': security_groups,
            'subnets': subnet_ids,
            's3_bucket': tier['resources']['drift.core.resources.awsdeploy']['s3_bucket'],
        },
        'deployable': {'deployable_name': conf.drift_app['name']},
        'apihandler': conf.drift_app.get('apihandler', 'drift.contrib.aws.lambdawsgi.handler'),
        'events': [],
        'log_forwarding_arn': log_forwarding_arn,
        'offline': offline,
    }

    if not template_args['apihandler'] and not template_args['events']:
        secho("Warning: Neither apihandler nor events defined. Nothing to do really.", fg='yellow')
        sys.exit(1)

    # Generate the serverless.yml configuration file
    secho("Generating serverless.yml")
    env = Environment(loader=FileSystemLoader(searchpath=os.path.dirname(__file__)))
    template = env.get_template('serverless.jinja.yml')
    serverless_yaml = template.render(**template_args)
    sls_config = yaml.load(serverless_yaml)

    if verbose:
        secho("\n-------- Template parameters: --------\n", bold=True)
        secho(pretty(template_args, 'json'))
        secho("\n-------- serverless.yml: --------\n", bold=True)
        secho(pretty(serverless_yaml, 'yaml'))

    with open(filename_yaml, 'w') as f:
        f.write(serverless_yaml)

    if preview:
        secho("Preview only. Exiting now.")
        sys.exit(1)

    _install_prerequisites(['serverless'], global_install=True)

    # Install plugins into this Serverless environment
    _install_prerequisites(sls_config['plugins'])

    if offline:
        sls_cmd = ['sls', 'offline', 'start']
    else:
        sls_cmd = ['sls', 'deploy']
        if verbose:
            sls_cmd += ['--verbose']

    echo("Running command: {}".format(' '.join(sls_cmd)))
    subprocess.call(sls_cmd)
    secho("Done in {:.0f} seconds!".format(time.time() - now), fg='green')


def _install_prerequisites(packages, global_install=False):

    echo("Installing packages: {}".format(packages))
    cmd = 'npm list --depth=0 --json'
    if global_install:
        cmd += ' -g'

    outs, errs = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE).communicate()
    npm_list = json.loads(outs.decode('ascii'))
    installed = list(npm_list.get('dependencies', {}).keys())

    for package_name in packages:
        if package_name not in installed:
            cmd = 'npm install ' + package_name
            if global_install:
                cmd = 'sudo {} -g'.format(cmd)
            secho("NPM package '{}' not installed. Executing: {}".format(package_name, cmd))
            ret = subprocess.call(cmd.split())
            if ret != 0:
                secho("Installation failed, try install the package manually.")
                sys.exit(ret)


if __name__ == '__main__':
    cli()



"""
Flask app with Datadog APM bootstrap.
Log level is set to 'INFO'.
"""
import os

from drift.flaskfactory import drift_app

if os.environ.get('ENABLE_DATADOG_APM', '0') == '1':
    import ddtrace
    ddtrace.patch_all()

app = drift_app()



# Deprecated
import warnings
from drift.contrib.flask.plainapp import app
warnings.warn(
    "Use drift.contrib.flask.plainapp instead of drift.uwsgiboot.",
    DeprecationWarning,
    stacklevel=2,
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import os
import logging
import random

# Note! boto3 is not imported here as this module is only used when configuring and provisioning
# resources on AWS. There is a big performance impact in importing boto3.

log = logging.getLogger(__name__)


# defaults when making a new tier
TIER_DEFAULTS = {
    "region": "<PLEASE FILL IN>",
    "ssh_key": "<PLEASE FILL IN>",
}


def register_resource_on_tier(ts, tier, attributes):
    """
    Tier registration callback.
    'tier' is from table 'tiers'.
    'attributes' is a dict containing optional attributes for default values.
    """
    # This resource requires AWS and can't run locally.
    if os.environ.get('DRIFT_USE_LOCAL_SERVERS', False):
        return

    import boto3

    # Provision a free-for-all S3 bucket for this tier.
    # Example: dgnorth.DEVNORTH or dgnorth.MYTIER.837
    bucket_tags = {'tier': tier['tier_name'], 'service-type': 'deployment'}
    buckets = find_resources(
        ['s3'],
        bucket_tags,
        region_name=attributes['region']
    )

    if buckets:
        # Bucket exists, all is fine.
        bucket_name = buckets[0]['arn'].replace('arn:aws:s3:::', '')
        log.info("S3 bucket for tier: %s", bucket_name)
    else:
        # Provision it
        bucket_name = '{}.{}'.format(ts.get_table('domain')['domain_name'], tier['tier_name'])
        bucket_name = bucket_name.lower()
        log.info("S3 bucket for tier not available. Creating one named '%s'", bucket_name)
        client = boto3.client('s3', region_name=attributes['region'])
        try:
            response = client.create_bucket(
                ACL='private',
                Bucket=bucket_name,
                CreateBucketConfiguration={'LocationConstraint': attributes['region']},
            )
        except Exception as e:
            if 'BucketAlreadyExists' not in str(e):
                raise

            bucket_name = '{}.{}'.format(bucket_name, random.randint(100, 999))
            response = client.create_bucket(
                ACL='private',
                Bucket=bucket_name,
                CreateBucketConfiguration={'LocationConstraint': attributes['region']},
            )

        def tag_resources(arns, tags, region_name=None):
            tags_client = boto3.client('resourcegroupstaggingapi', region_name=region_name)
            tags_client.tag_resources(ResourceARNList=arns, Tags=tags)

        tag_resources(['arn:aws:s3:::' + bucket_name], bucket_tags, attributes['region'])
        log.info("S3 bucket for tier created: %s", bucket_name)
        log.info("Response: %s", response)

    attributes['s3_bucket'] = bucket_name

    # LEGACY SUPPORT! Copy the 'attributes' to the root as 'aws':
    tier['aws'] = attributes


def find_resources(recource_types, tags=None, region_name=None):
    """
    Find all resources with the matching resource types and tags.

    For resource types and ARN formats see:
    https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html

    'recource_types' is a list of service names or service:type names ('s3', 'ec2:instance, ...')
    'tags' is a dict of tag:value.

    Returns list of dict with 'arn' and 'tags'.
    """
    import boto3

    ret = []
    tags_client = boto3.client('resourcegroupstaggingapi', region_name=region_name)
    paginator = tags_client.get_paginator('get_resources')
    if tags:
        tags = [{'Key': key, 'Values': [str(value)]} for key, value in tags.items()]
    else:
        tags = []
    for page in paginator.paginate(TagFilters=tags, ResourceTypeFilters=recource_types):
        for resource in page['ResourceTagMappingList']:
            entry = {
                'arn': resource['ResourceARN'],
                'tags': {tag['Key']: tag['Value'] for tag in resource['Tags']}
            }
            ret.append(entry)

    return ret





"""
Sentry integration
"""
import logging
import os, sys

from flask import current_app, g
from driftconfig.util import get_drift_config
from drift.utils import get_tier_name
import sentry_sdk
from sentry_sdk.integrations.flask import FlaskIntegration
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
from sentry_sdk.integrations.redis import RedisIntegration
from drift.core.extensions.logging import get_log_details

log = logging.getLogger(__name__)


TIER_DEFAULTS = {"dsn": "<PLEASE FILL IN>"}


# Initialize Sentry at file scope to catch'em all.
def _init_sentry(app):
    dsn = os.environ.get('SENTRY_DSN')
    tier_name = get_tier_name()
    conf = get_drift_config(tier_name=tier_name)

    if not dsn:
        tier = conf.tier
        if tier and 'drift.core.resources.sentry' in tier['resources']:
            sentry_config = tier['resources']['drift.core.resources.sentry']
            dsn = sentry_config.get('dsn')
        if dsn == TIER_DEFAULTS['dsn']:
            # Configuration value not set yet
            dsn = None

    if not dsn:
        log.info("Sentry is not enabled")
        return

    sentry_sdk.init(
        dsn=dsn,
        integrations=[
            FlaskIntegration(),
            LoggingIntegration(event_level=None, level=None),
        ],
        environment=tier_name,
    )
    integrations = []

    if "drift.core.resources.redis" in app.config["resources"]:
        integrations.append(RedisIntegration())

    if "drift.core.resources.postgres" in app.config["resources"]:
        integrations.append(SqlalchemyIntegration())

    sentry_sdk.init(
        dsn=dsn,
        integrations=integrations,
        before_send=before_send,
        environment=get_tier_name(),
    )

    return True


def before_send(event, hint):
    details = get_log_details()

    """
    details["logger"] =
    {
      "created": "2019-11-19T11:26:58.019125Z",
      "remote_addr": "127.0.0.1",
      "tenant": "acme-prod-test",
      "tier": "TIER"
    }

    details["user"] =
    {
      "client_id": 13,
      "identity_id": 200000011,
      "jti": "pxS59dsk98uSpo7Ee9Ci",
      "player_id": 11,
      "player_name": "",
      "roles": [],
      "user_id": 100000011,
      "user_name": "28f28471"
    }
    """
    user = {}
    if "remote_addr" in details["logger"]:
        user["ip_address"] = details["logger"]["remote_addr"]

    if "user" in details:
        user["id"] = details["user"]["user_id"]
        user["username"] = details["user"]["user_name"]

    if user:
        event.setdefault("user", {}).update(user)

    event.setdefault("extra", {}).update(details)

    if details["logger"]["tenant"]:
        event.setdefault("tags", {})["tenant"] = details["logger"]["tenant"]

    return event


def drift_init_extension(app, **kwargs):
    if not _init_sentry(app):
        log.warning(
            "Sentry not initialized. run 'driftconfig assign-tier' to refresh the config."
        )


def register_deployable(ts, deployablename, attributes):
    """
    Deployable registration callback.
    'deployablename' is from table 'deployable-names'.
    """
    pass


def register_deployable_on_tier(ts, deployable, attributes):
    """
    Deployable registration callback for tier.
    'deployable' is from table 'deployables'.
    """
    pass


def register_resource_on_tier(ts, tier, attributes):
    """
    Tier registration callback.
    'tier' is from table 'tiers'.
    'attributes' is a dict containing optional attributes for default values.
    """
    pass


def register_deployable_on_tenant(
    ts, deployable_name, tier_name, tenant_name, resource_attributes
):
    pass


def log_sentry(msg, *args, **kwargs):
    """Write a custom 'error' log event to sentry. Behaves like normal loggers
    Batches messages that have the same signature. Please use '"hello %s", "world"', not '"hello %s" % "world"'
    Example usage: log_sentry("Hello %s", "world", extra={"something": "else"})
    """
    try:
        # also log out an error for good measure
        log.error(msg, *args)
        if not current_app:
            return

        extra = kwargs.get('extra', {})
        if getattr(g, 'log_defaults', None):
            extra.update(g.log_defaults)
        # get info on the caller
        f_code = sys._getframe().f_back.f_code
        extra['method'] = f_code.co_name
        extra['line'] = f_code.co_firstlineno
        extra['file'] = f_code.co_filename

        with sentry_sdk.push_scope() as scope:
            for k, v in extra.items():
                scope.set_extra(k, v)
            scope.level = 'warning'
            sentry_sdk.capture_message(msg % args)

    except Exception as e:
        log.exception(e)



"""
Deploy a local build to the running cluster.
Note that this is currently just calling the old setup.py script
and will be refactored soon.
"""
import os
import os.path
import subprocess
import sys
from fabric import Connection, Config
import tempfile
import shutil
import pkg_resources
import time

import requests
from click import echo, secho
from six import print_, StringIO

from drift.management import get_ec2_instances, get_app_version
from drift.utils import get_config

EC2_USERNAME = 'ubuntu'
UWSGI_LOGFILE = "/var/log/uwsgi/uwsgi.log"


def get_options(parser):
    parser.add_argument(
        "--ip",
        help="Deploy to a certain instance instead of across the cluster")
    parser.add_argument(
        "--drift",
        help="Force deploy the drift library from the local filesystem (hack)",
        action="store_true")
    parser.add_argument(
        "-c",
        "--comment",
        help="Short description of the changes",
        default=None)
    parser.add_argument(
        "--deploy-to-this-tier", dest='tiername',
        help="Use to override tier protection. State the name of the expected tier.",
        default=None)
    parser.add_argument(
        "--force-reinstall", dest='forcereinstall',
        help="Run pipenv install",
        action="store_true")


def run_command(args):
    t = time.time()
    conf = get_config()
    if not conf.deployable:
        echo("Deployable '{}' not found in config '{}'.".format(
            conf.drift_app['name'], conf.domain['domain_name']))
        sys.exit(1)

    service_name = conf.deployable['deployable_name']
    tier = conf.tier['tier_name']
    region = conf.tier['aws']['region']
    ssh_key_name = conf.tier['aws']['ssh_key']
    ssh_key_file = os.path.expanduser('~/.ssh/{}.pem'.format(ssh_key_name))

    include_drift = args.drift

    if args.tiername and args.tiername != tier:
        echo("Default tier is '{}' but you expected '{}'. Quitting now.".format(tier, args.tiername))
        return

    if conf.tier['is_live'] and tier != args.tiername:
        echo("You are quickdeploying to '{}' which is a protected tier.".format(tier))
        echo("This is not recommended!")
        echo("If you must do this, and you know what you are doing, state the name of")
        echo("the tier using the --deploy-to-this-tier argument and run again.")
        return

    # The idea:
    # - Generate a source distribution of the current project and all other projects that are
    #   referenced (drift and drift-config in particular).
    # - Upload the zip files to ~/ at the remote host.
    # - Unzip the files into app root.
    # - Restart the uwsgi service.

    project_folders = ["."]
    if include_drift:
        import drift
        drift_folder = os.path.abspath(os.path.join(drift.__file__, '..', '..'))
        project_folders.append(drift_folder)

    def deploy(distros):
        shell_scripts = []

        for project_folder in project_folders:
            echo("Creating source distribution from {!r}".format(project_folder))
            # TODO: This code mirrors the one in ami.py. It's not DRY.
            cmd = [
                sys.executable,  # "python",
                os.path.join(project_folder, "setup.py"),
                "sdist",
                "--formats=tar",
                "--dist-dir=" + distros,
            ]

            p = subprocess.Popen(
                cmd, cwd=project_folder, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            stdout, _ = p.communicate()
            stdout = str(stdout.decode("utf-8"))  # while still running py2
            if p.returncode != 0:
                print_(stdout)
                sys.exit(p.returncode)

            # Use custom quickdeploy script if found.
            # Prefix them with environment variables:
            header = "#!/bin/bash\n"
            # Minor hacketyhack
            header += "export service={}\n".format(service_name)
            header += "export version={}\n".format(get_app_version())

            header += "export UWSGI_LOGFILE={}\n\n".format(UWSGI_LOGFILE)
            if not args.forcereinstall:
                header += "export SKIP_PIP=1\n"

            quickdeploy_script_file = os.path.join(project_folder, "scripts/quickdeploy.sh")
            if os.path.exists(quickdeploy_script_file):
                echo("Using quickdeploy.sh from this project.")
            else:
                echo("Using standard quickdeploy.sh from Drift library")
                # Use standard quickdeploy script. Only works for web stacks.
                quickdeploy_script_file = pkg_resources.resource_filename(__name__, "quickdeploy.sh")
            with open(quickdeploy_script_file, 'r') as f:
                src = header + f.read().replace("#!/bin/bash", "")
                shell_scripts.append(src)

        for ec2 in get_ec2_instances(region, tier, service_name):
            if args.ip and ec2.private_ip_address != args.ip:
                echo("Skipping {!r}".format(ec2.private_ip_address))
                continue

            conf = Config()
            conf.run.warn = True
            conf.connect_kwargs.key_filename = ssh_key_file
            conn = Connection(host=ec2.private_ip_address, user=EC2_USERNAME, config=conf)

            for dist_file in os.listdir(distros):
                echo("Installing {} on {}".format(dist_file, ec2.private_ip_address))
                full_name = os.path.join(distros, dist_file)

                # Remove the previous file forcefully, if needed
                conn.run("sudo rm -f {}".format(dist_file))
                conn.put(full_name)

            echo("Running quickdeploy script on {}".format(ec2.private_ip_address))
            for shell_script in shell_scripts:
                temp = '/tmp/quickdeploy.sh'  # could use tmpname on the remote...
                conn.put(StringIO(shell_script), temp)
                conn.sudo('bash "{}"'.format(temp))
                conn.run('rm -f "{}"'.format(temp))

            # See if the server responds to an http request
            secho("Pinging endpoint:", bold=True)
            timeout = 5.0
            retries = 15
            wait_between = 0.5
            for i in range(retries):
                try:
                    ret = requests.get(
                        'http://{}:8080'.format(ec2.private_ip_address),
                        timeout=timeout,
                        )
                except Exception as e:
                    if 'Read timed out' in str(e):
                        secho("WARNING! Web server timeout in {} seconds.".format(timeout), fg="yellow")
                    elif "Max retries exceeded" in str(e):
                        pass
                    else:
                        secho("ERROR! {}".format(e), fg="red")
                else:
                    secho("SUCCESS: Instance {}  is serving. status code: {}".format(ec2.private_ip_address, ret.status_code), fg="green")
                    break

                if i < retries - 1:
                    secho("Endpoint not responding, retrying in {} seconds...".format(wait_between))
                    time.sleep(wait_between)
            else:
                secho("ERROR: Instance is not responding!", fg='red')

            # todo: see if this needs to be done as well:
            # _set_ec2_tags(ec2, deployment_manifest, "drift:manifest:")

    # Wrap the business logic in RAII block
    distros = tempfile.mkdtemp(prefix='drift.quickdeploy.')
    try:
        deploy(distros)
    finally:
        shutil.rmtree(distros)

    echo("Quickdeploy ran for {:.1f} seconds.".format(time.time() - t))



# -*- coding: utf-8 -*-
import os
import os.path
import logging
from functools import wraps
from socket import gethostname
import uuid
import json
import six
import textwrap

# pygments is optional for now
try:
    got_pygments = True
    from pygments import highlight, util
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import get_formatter_by_name, get_all_formatters
    from pygments.styles import get_style_by_name, get_all_styles
except ImportError:
    got_pygments = False

from flask import g, make_response, jsonify, request, current_app, url_for
from click import echo
from flask_marshmallow.fields import AbsoluteURLFor
import flask_smorest

from driftconfig.util import get_drift_config

from drift.core.extensions.tenancy import tenant_from_hostname

log = logging.getLogger(__name__)

host_name = gethostname()


def get_app_root():
    """Returns absolute path to the current application root directory."""
    return os.path.expanduser(os.environ.get('DRIFT_APP_ROOT', os.path.abspath('.')))


def get_config(ts=None, tier_name=None, tenant_name=None):
    """Wraps get_drift_config() by providing default values for tier, tenant and drift_app."""
    # Hack: Must delay import this
    # TODO: Stop using this function. Who is doing it anyways?
    from drift.flaskfactory import load_flask_config
    if current_app:
        app_ts = current_app.extensions['driftconfig'].table_store
        if ts is not app_ts:
            log.warning("Mismatching table_store objects in get_config(): ts=%s, app ts=%s", ts, app_ts)
        ts = app_ts

    conf = get_drift_config(
        ts=ts,
        tier_name=tier_name or get_tier_name(),
        tenant_name=tenant_name or tenant_from_hostname,
        drift_app=load_flask_config(),
    )
    return conf


def get_tenant_name():
    """
    Return the current tenant name.
    If inside a Flask request context, it's the one defined by that context,
    and if not, then it must be specified explicitly in the environment
    variable 'DRIFT_DEFAULT_TENANT'.
    """
    if g and hasattr(g, 'conf'):
        return g.conf.tenant['tenant_name']
    elif 'DRIFT_DEFAULT_TENANT' in os.environ:
        return os.environ['DRIFT_DEFAULT_TENANT']
    else:
        raise RuntimeError(
            "No default tenant available in this context. Specify one in "
            "'DRIFT_DEFAULT_TENANT' environment variable, or use the --tenant command "
            "line argument."
        )


def uuid_string():
    return str(uuid.uuid4()).split("-")[0]


def is_ec2():
    """Naive check if this is an ec2 instance"""
    return host_name and host_name.startswith("ip")


def json_response(message, status=200, fields=None):
    d = {
        "message": message,
        "status": status
    }
    if fields:
        d.update(fields)
    log.info("Generated json response %s : %s", status, message)
    return make_response(jsonify(d), status)


def client_debug_message(message):
    """write a message to the response header for consumption by the client.
    Used the Drift-Debug-Message header"""
    g.client_debug_messages.append(message)


def validate_json(required):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kw):
            try:
                request.json
            except Exception:
                return json_response(
                    "This endpoint requires a json request body", 400
                )
            for r in required.split(","):
                if r not in (request.json or {}):
                    log.warning(
                        "Required field not specified: %s, json is %s",
                        r, request.json
                    )
                    return make_response(jsonify(
                        {
                            "message": "Required field not specified: %s" % r,
                            "status": 500
                        }), 500)

            return f(*args, **kw)
        return wrapper
    return decorator


def add_response_headers(headers={}):
    """This decorator adds the headers passed in to the response"""
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            resp = make_response(f(*args, **kwargs))
            h = resp.headers
            for header, value in headers.items():
                h[header] = value
            return resp
        return decorated_function
    return decorator


def get_tier_name(fail_hard=True):
    """
    Get tier name from environment
    """
    if 'DRIFT_TIER' in os.environ:
        return os.environ['DRIFT_TIER']

    if fail_hard:
        raise RuntimeError(
            "No tier specified. Specify one in "
            "'DRIFT_TIER' environment variable, or use the --tier command "
            "line argument."
        )


def request_wants_json():
    """
    Returns true it the request header has 'application/json'.
    This is used to determine whether to return html or json content from
    the same endpoint
    """
    best = request.accept_mimetypes \
        .best_match(['application/json', 'text/html'])
    return best == 'application/json' and \
        request.accept_mimetypes[best] > \
        request.accept_mimetypes['text/html']


PRETTY_FORMATTER = 'console256'
PRETTY_STYLE = 'tango'


def pretty(ob, lexer=None):
    """
    Return a pretty console text representation of 'ob'.
    If 'ob' is something else than plain text, specify it in 'lexer'.

    If 'ob' is not string, Json lexer is assumed.

    If 'pretty.unicorns' are enabled, the world is a happier place.

    Command line switches can be used to control highlighting and style.
    """
    if lexer is None:
        if isinstance(ob, six.string_types):
            lexer = 'text'
        else:
            lexer = 'json'

    if lexer == 'json':
        ob = json.dumps(ob, indent=4, sort_keys=True)

    if got_pygments:
        lexerob = get_lexer_by_name(lexer)
        formatter = get_formatter_by_name(PRETTY_FORMATTER, style=PRETTY_STYLE)
        # from pygments.filters import *
        # lexerob.add_filter(VisibleWhitespaceFilter(spaces=True, tabs=True, newlines=True))
        ret = highlight(ob, lexerob, formatter)
    else:
        if pretty.unicorns:
            pretty.unicorns = False
            print(textwrap.dedent("""\


                Note! All this blurb would look much better with colors!.
                "Plese Run the following command for the sake of rainbows and unicorns:
                "pip install pygments

                """))

        ret = ob

    return ret.rstrip()


pretty.unicorns = True  # Always tell the world about how pretty it can be


def set_pretty_settings(formatter=None, style=None):
    if not got_pygments:
        return

    global PRETTY_FORMATTER
    global PRETTY_STYLE

    try:
        if formatter:
            get_formatter_by_name(formatter)
            PRETTY_FORMATTER = formatter

        if style:
            get_style_by_name(style)
            PRETTY_STYLE = style

    except util.ClassNotFound as e:
        echo("Note: {}".format(e))
        echo(get_avaible_pretty_settings())


def get_avaible_pretty_settings():
    formatters = ', '.join([f.aliases[0] for f in get_all_formatters()])
    styles = ', '.join(list(get_all_styles()))
    s = "Available formatters: {}\nAvailable styles: {}".format(formatters, styles)
    return s


class Url(AbsoluteURLFor):
    """
    Extends AbsoluteURLFor field. If 'kwargs' contains 'doc' it will not be passed to
    AbsoluteURLFor but instead added to the metadata of the parent class as 'description'.
    """
    def __init__(self, endpoint, **kwargs):
        doc = kwargs.pop('doc')
        super().__init__(endpoint, **kwargs)
        if doc:
            self.metadata['description'] = doc


def set_url_in_header(location_url):
    response_header = {"Location": location_url}
    flask_smorest.utils.get_appcontext()['headers'].update(response_header)


def set_url_for_in_header(*args, **kw):
    kw['_external'] = True
    set_url_in_header(url_for(*args, **kw))




import unittest
import logging

from flask import Flask


@unittest.skip("needs refactoring")
class driftTestCase(unittest.TestCase):

    def setUp(self):
        self.app = Flask(__name__)
        logging.basicConfig(level="ERROR")
        self.app.testing = True
        self.test_client = self.app.test_client()

    def tearDown(self):
        pass

    def test_flasksetup(self):
        # Run minimal setup
        # flasksetup(self.app, options=[])
        pass

    def test_all(self):
        # Run with all options
        # flasksetup(self.app)
        pass


if __name__ == "__main__":

    unittest.main()



# -*- coding: utf-8 -*-
import unittest

from six.moves import http_client
from flask import Flask

from drift.core.extensions.httpmethod import drift_init_extension


class HTTPMethodTestCase(unittest.TestCase):

    def setUp(self):
        app = Flask(__name__)
        self.app = app
        app.config['TESTING'] = True

        @app.route('/some-endpoint', methods=['PATCH'])
        def some_endpoint():
            return 'success'

    def test_httpmethod(self):
        with self.app.test_client() as c:
            # Try and fail to access a PATCH endpoint using GET.
            resp = c.get('/some-endpoint')
            self.assertEqual(resp.status_code, http_client.METHOD_NOT_ALLOWED)

            # Try and fail to access a PATCH endpoint using GET and override, but without
            # the handler installed.
            resp = c.get('/some-endpoint', headers={'X-HTTP-Method-Override': 'PATCH'})
            self.assertEqual(resp.status_code, http_client.METHOD_NOT_ALLOWED)

            # Install the handler, then try and succeed to access a PATCH endpoint
            # using GET and override.
            drift_init_extension(self.app, api=None)
            resp = c.get('/some-endpoint', headers={'X-HTTP-Method-Override': 'PATCH'})
            self.assertEqual(resp.status_code, http_client.OK)
            self.assertEqual(resp.data.decode("ascii"), 'success')


if __name__ == '__main__':
    unittest.main()



import os
import os.path
import sys

from fabric import Connection, Config
from click import echo, secho

from drift.management import get_ec2_instances
from drift.utils import get_config

EC2_USERNAME = 'ubuntu'
UWSGI_LOGFILE = "/var/log/uwsgi/uwsgi.log"


def get_options(parser):
    parser.add_argument(
        "--ip",
        help="Deploy to a certain instance instead of across the cluster")
    parser.add_argument(
        "--stream",
        help="Stream live logs from the host (requires the --host argument)",
        action="store_true")
    parser.add_argument(
        "--host",
        help="Pick a single host to view logs from",
        default=None)
    parser.add_argument(
        "--grep",
        help="Grep pattern to search for",
        default=None)


def run_command(args):
    conf = get_config()
    if not conf.deployable:
        secho("Deployable '{}' not found in config '{}'.".format(
            conf.drift_app['name'], conf.domain['domain_name']), fg="red")
        sys.exit(1)

    deployable_name = conf.deployable['deployable_name']
    tier = conf.tier['tier_name']
    region = conf.tier['aws']['region']
    ssh_key_name = conf.tier['aws']['ssh_key']
    ssh_key_file = os.path.expanduser('~/.ssh/{}.pem'.format(ssh_key_name))

    echo()
    echo("*** VIEWING LOGS FOR SERVICE '{}' / TIER '{}' IN REGION '{}'\n".format(deployable_name, tier, region))

    instances = get_ec2_instances(region, tier, deployable_name)

    if args.host:
        instances = [i for i in instances if i.private_ip_address == args.host]
    echo("Gathering logs from '%s' on the following instances:" % UWSGI_LOGFILE)
    for inst in instances:
        echo("   %s" % inst.private_ip_address)

    if args.stream and len(instances) > 1:
        secho("The --stream argument can only be used on a single host. Please use --host to pick one", fg="red")
        return

    for ec2 in instances:
        ip_address = ec2.private_ip_address
        echo("*** Logs in {} on {}...".format(UWSGI_LOGFILE, ip_address))
        if not args.stream:
            conf = Config()
            conf.connect_kwargs.key_filename = ssh_key_file
            conn = Connection(host=ip_address, user=EC2_USERNAME, config=conf)
            cmd = "sudo tail {} -n 100".format(UWSGI_LOGFILE)
            if args.grep:
                cmd += " | grep {}".format(args.grep)
            conn.run(cmd)
            echo()
        else:
            import paramiko
            import select

            client = paramiko.SSHClient()
            client.load_system_host_keys()
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            key_path = os.path.expanduser(ssh_key_file)
            client.connect(ip_address, username=EC2_USERNAME, key_filename=key_path)
            channel = client.get_transport().open_session()
            grep_cmd = ""
            if args.grep:
                grep_cmd = " | grep --line-buffered {}".format(args.grep)
            channel.exec_command("sudo tail -f {}{}".format(UWSGI_LOGFILE, grep_cmd))
            while True:
                if channel.exit_status_ready():
                    break
                rl, wl, xl = select.select([channel], [], [], 0.0)
                if len(rl) > 0:
                    sys.stdout.write(channel.recv(1024))



# -*- coding: utf-8 -*-

import logging

log = logging.getLogger(__name__)

DEFAULT_PORT = 5000


def run_app(app, web_server=None):
    """
    Run Flask app in the specified web server.
    :param app: an instance of :class:`flask.Flask`.
    :param web_server: is one of:
        "flask":    Use Flask default web server (Werkzeug).
        "tornado":  Use Tornado web server.
        "gevent":   Use Gevent web server.
        "twisted":  Use Twisted Web web server.
        "gunicorn": Use Gunicorn web server.
        "mod_wsgi": Use Apache web server.
        "uwsgi":    Use uWSGI web server.

    If :param web_server: is None, the "web_server" config value from `app`
    is used.

    Note: Only "flask", "tornado", "gevent", and "twisted" work a the moment.
    """
    web_server = web_server or "flask"
    log.info("Running '%s' web server on port %s", web_server, app.config.get('PORT', DEFAULT_PORT))

    # Magically map 'web_server' to a local function name.
    import webservers
    fn = getattr(webservers, "run_{}_server".format(web_server), None)
    if fn is None:
        raise RuntimeError("Web server '%s' not supported." % web_server)
    fn(app, app.config.get('PORT', DEFAULT_PORT))


def run_flask_server(app, port):
    app.run(threaded=True, host='0.0.0.0', port=port)


def run_tornado_server(app, port):

    from tornado.wsgi import WSGIContainer
    from tornado.httpserver import HTTPServer
    from tornado.ioloop import IOLoop
    log.info("Activating Tornado server on port %s", port)
    http_server = HTTPServer(WSGIContainer(app), xheaders=True)
    http_server.listen(port)
    IOLoop.instance().start()


def run_gevent_server(app, port):
    from gevent.pywsgi import WSGIServer
    import gevent.monkey
    gevent.monkey.patch_all()

    log.info("Activating gevent server on port %s", port)
    http_server = WSGIServer(('', port), app)
    http_server.serve_forever()


def run_twisted_server(app, port):
    import sys

    try:
        import _preamble
        _preamble
    except ImportError:
        sys.exc_clear()

    sys.argv = ['runserver.py', 'web', '--port',
                str(port), '--wsgi', 'app.app']
    from twisted.scripts.twistd import run
    run()


def run_gunicorn_server(app, port):
    # this is not tested and does not work. only runs on linux.
    raise RuntimeError("Not Implemented yet!")
    import sys
    from pkg_resources import load_entry_point

    sys.exit(
        load_entry_point(
            'gunicorn==19.0.0', 'console_scripts', 'gunicorn')()
    )


def run_mod_wsgi_server(app, port):
    raise RuntimeError("Not Implemented yet!")


def run_uwsgi_server(app, port):
    raise RuntimeError("Not Implemented yet!")
    # this is not tested and does not work. only runs on linux.
    import sys

    sys.argv = ['runserver.py', '-s', '/tmp/uwsgi.sock', '-w', 'app.app']



# -*- coding: utf-8 -*-
import warnings

warnings.warn(
    "please import drift.core.extensions.urlregistry instead of drift.urlregistry.",
    DeprecationWarning,
    stacklevel=2)


# HACK: Temporary placeholder file until billion imports have been fixed up
def register_endpoints(f):
    from drift.core.extensions.urlregistry import register_endpoints as reg
    return reg(f)



from click import echo, secho
from alembic import context
from sqlalchemy import pool, create_engine

from drift.flaskfactory import drift_app
from drift.core.resources.postgres import format_connection_string, connect
from drift.utils import get_tier_name, get_config
from drift.orm import Base


MASTER_USERNAME = 'postgres'
MASTER_PASSWORD = 'postgres'


def run_migrations():

    conf = get_config()
    pick_tenant = context.get_x_argument(as_dictionary=True).get('tenant')
    dry_run = context.get_x_argument(as_dictionary=True).get('dry-run')

    for tenant in conf.tenants:
        tenant_name = tenant['tenant_name']
        secho("Tenant '{}': ".format(tenant_name), nl=False)
        pginfo = tenant.get('postgres')
        if not pginfo:
            secho("Missing postgres resource info!", fg='red')
            continue

        if pick_tenant and tenant_name != pick_tenant:
            secho("Skipping this tenant.", fg='yellow')
            continue

        if dry_run:
            secho("Dry run, not taking any further actions.")

        if context.is_offline_mode():
            sql_filename = '{}.{}.sql'.format(conf.tier['tier_name'], tenant_name)
            echo("Writing SQL code to ", nl=False)
            secho(sql_filename, fg='magenta')
            with open(sql_filename, 'w') as out:
                context.configure(
                    url=format_connection_string(pginfo),
                    output_buffer=out,
                    target_metadata=Base.metadata,
                    process_revision_directives=process_revision_directives,
                    compare_type=True,
                )

                with context.begin_transaction():
                    context.run_migrations()
        else:
            # HACK WARNING! This has to come from the config
            pginfo['username'] = MASTER_USERNAME
            pginfo['password'] = MASTER_PASSWORD

            secho("\n\tConnecting {server}:{port}/{database}...".format(**pginfo), nl=False)

            engine = connect(pginfo, connect_timeout=3.0)
            try:
                connection = engine.connect()
            except Exception as e:
                if 'timeout expired' in str(e):
                    secho("ERROR: {}".format(e), fg='red')
                    continue
                else:
                    raise

            transaction = connection.begin()
            secho("OK", fg="green")
            secho("\tRunning migration...", nl=False)
            context.configure(
                connection=connection,
                upgrade_token="%s_upgrades" % tenant_name,
                downgrade_token="%s_downgrades" % tenant_name,
                target_metadata=Base.metadata,
            )
            context.run_migrations()
            transaction.commit()
            connection.close()
            secho("OK", fg="green")


def process_revision_directives(context, revision, directives):
    if context.config.cmd_opts.autogenerate:
        script = directives[0]
        if script.upgrade_ops.is_empty():
            directives[:] = []


def run():

    app = drift_app()
    run_migrations()
    del app



# -*- coding: utf-8 -*-
"""
JWT Session Management

By including this resource module in a Drift app, it will be able to accept JWT from a list
of trusted issuers and issue and sign new JWT's.

Custom attributes for top level registration:

    key_size:            <int>  Size in bytes of private key.
    trusted_issuers:     <list of deployable names>  Default value is ['drift-base']
    expiry_days:         <int>  Expiration in days, default is 365.
"""
import logging
import datetime
import secrets

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend

DEFAULT_KEY_SIZE = 1024
DEFAULT_EXPIRY_DAYS = 365

log = logging.getLogger(__name__)


def register_deployable(ts, deployablename, attributes):
    """
    Deployable registration callback.
    'deployablename' is from table 'deployable-names'.
    """
    pass


def register_deployable_on_tier(ts, deployable, attributes):
    """
    Deployable registration callback for tier.
    'deployable' is from table 'deployables'.
    """
    pk = {'tier_name': deployable['tier_name'], 'deployable_name': deployable['deployable_name']}
    row = ts.get_table('public-keys').get(pk)
    if row is None:
        row = ts.get_table('public-keys').add(pk)

    # Add session cookie secret key
    row.setdefault('secret_key', secrets.token_urlsafe(32))

    # Generate RSA key pairs
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=attributes.get('key_size', DEFAULT_KEY_SIZE),
        backend=default_backend()
    )

    public_key = private_key.public_key()

    private_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()
    )

    public_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )

    now = datetime.datetime.utcnow()
    expiry_days = attributes.get('expiry_days', DEFAULT_EXPIRY_DAYS)

    keypair = {
        'issued': now.isoformat() + "Z",
        'expires': (now + datetime.timedelta(days=expiry_days)).isoformat() + "Z",
        'public_key': public_pem.decode(),  # PEM is actually a text format
        'private_key': private_pem.decode(),  # PEM is actually a text format
    }

    # LEGACY SUPPORT! Make sure there is only one keypair registered. If one exists already
    # in the config then leave it as is. If not, add one.
    if 'keys' in row and len(row['keys']) > 0:
        # Just leave this as is
        log.warning("Legacy support: Key pair already registered, leaving it as is.")
        current_keypair = row['keys'][0]
    else:
        log.warning("Legacy support: Adding new key pair for this deployable.")
        current_keypair = keypair
        row.setdefault('keys', []).append(keypair)

    # LEGACY SUPPORT! Register drift-base as trusted issuer. Always.
    if deployable['deployable_name'] == 'drift-base':
        issuers = deployable.setdefault('jwt_trusted_issuers', [])
        for issuer in issuers:
            if issuer.get('iss') == 'drift-base':
                log.warning("Legacy support: drift-base already configured as trusted issuer.")
                break
        else:
            log.warning("Legacy support: Adding drift-base as trusted issuer.")
            issuers.append({
                'iss': 'drift-base',
                'iat': current_keypair['issued'],
                'exp': current_keypair['expires'],
                'pub_rsa': current_keypair['public_key'],
            })


def register_resource_on_tier(ts, tier, attributes):
    """
    Tier registration callback.
    'tier' is from table 'tiers'.
    'attributes' is a dict containing optional attributes for default values.
    """
    # LEGACY SUPPORT! Register the service user
    tier.setdefault('service_user', {"password": "SERVICE", "username": "user+pass:$SERVICE$"})


def register_deployable_on_tenant(ts, deployable_name, tier_name, tenant_name, resource_attributes):
    pass



# -*- coding: utf-8 -*-
import os
import sys
import uuid
import copy
import unittest
import responses
import requests
import re
import jwt

from six.moves import http_client

from driftconfig.util import set_sticky_config, get_default_drift_config
import driftconfig.testhelpers

from drift.core.extensions.jwt import JWT_ALGORITHM, register_auth_provider
from .flaskfactory import drift_app


import logging
log = logging.getLogger(__name__)

service_username = "user+pass:$SERVICE$"
service_password = "SERVICE"
local_password = "LOCAL"
AUTH_TEST_PROVIDER = 'unit_test'

big_number = 9999999999


def uuid_string():
    return str(uuid.uuid4()).split("-")[0]


def make_unique(name):
    """Make 'name' unique by appending random numbers to it."""
    return name + uuid_string()


db_name = None


def flushwrite(text):
    sys.stdout.write(text + '\n')
    sys.stdout.flush()


_tenant_is_set_up = False


def setup_tenant(config_size=None, use_app_config=True):
    """
    Called from individual test modules.
    create a tenant only if the test module was not called from
    the kitrun's systest command
    (in which case drift_test_database has been set in environ)
    Also configure some basic parameters in the app.

    TODO: Fix up app config. In the meantime:
    If 'use_app_config' is True, the tenant will be set up using
    app config from /config/config.json. If not, the tenant is set
    up using empty app config.

    Returns the table store object for the current config.
    """
    global _tenant_is_set_up
    if _tenant_is_set_up:
        return get_default_drift_config()

    _tenant_is_set_up = True

    # TODO: Fix this app config business
    from drift.flaskfactory import load_flask_config, set_sticky_app_config
    if use_app_config:
        app_config = load_flask_config()
    else:
        app_config = {'resources': [], 'resource_attributes': {}}
    set_sticky_app_config(app_config)

    # Create a test tenant, including the kitchen sink.
    ts = driftconfig.testhelpers.create_test_domain(
        config_size=config_size,
        resources=app_config['resources'],
        resource_attributes=app_config['resource_attributes'],
    )
    set_sticky_config(ts)

    # Pick any tier as the default tier then pick any tenant and deployable from that tier
    # as the default for each respectively.
    tier = ts.get_table('tiers').find()[0]
    tier_name = tier['tier_name']
    tenant = ts.get_table('tenants').find({'tier_name': tier_name})[0]
    deployable = ts.get_table('deployables').find({'tier_name': tier_name})[0]

    os.environ['DRIFT_TIER'] = tier_name
    os.environ['DRIFT_DEFAULT_TENANT'] = tenant['tenant_name']
    app_config['name'] = deployable['deployable_name']

    # mixamix
    app_config['TESTING'] = True

    return ts


def remove_tenant():
    """
    Called from individual test modules.
    """
    global _tenant_is_set_up
    if _tenant_is_set_up:
        driftconfig.testhelpers.terminate_tenants(get_default_drift_config())
        from drift.flaskfactory import set_sticky_app_config
        set_sticky_app_config(None)
        set_sticky_config(None)
        _tenant_is_set_up = False


class DriftBaseTestCase(unittest.TestCase):
    headers = {}
    token = None
    current_user = {}
    endpoints = {}
    player_id = None
    user_id = None
    player_name = None
    auth_provider = AUTH_TEST_PROVIDER

    @staticmethod
    def mock(func):
        @responses.activate
        def wrapped(self, *args, **kwargs):
            self._setup_mocking()
            return func(self, *args, **kwargs)

        return wrapped

    def _do_request(self, method, endpoint, data=None,
                    params=None, *args, **kw):

        """
        Note that here we must use a inner function, otherwise mock
        will be evaluated at the module import time.
        """
        @DriftBaseTestCase.mock
        def inner(self, method, endpoint, data, params, *args, **kw):
            check = kw.pop("check", True)
            expected_status_code = kw.pop("expected_status_code", http_client.OK)
            headers = copy.copy(self.headers)
            if "Accept" not in headers:
                headers["Accept"] = "application/json"

            if data:
                headers["Content-Type"] = "application/json"
                if not isinstance(data, list) and not isinstance(data, dict):
                    raise Exception("Data must be a list or a dict: %s" % data)
            if db_name:
                headers["tenant"] = db_name

            if not endpoint.startswith(self.host):
                endpoint = self.host + endpoint

            r = getattr(requests, method)(
                endpoint,
                json=data,
                headers=headers,
                params=params

            )
            if check:
                self.assertEqual(
                    r.status_code, expected_status_code,
                    u"Status code should be {} but is {}: {}".format(
                        expected_status_code,
                        r.status_code, r.text.replace("\\n", "\n")
                    )
                )
            return r
        return inner(self, method, endpoint, data, params, *args, **kw)

    def _setup_mocking(self):
        def _mock_callback(request):
            method = request.method.lower()
            url = request.path_url
            handler = getattr(self.app, method)
            r = handler(
                url,
                data=request.body,
                headers=dict(request.headers)
            )
            return (r.status_code, r.headers, r.data)

        pattern = re.compile("{}/(.*)".format(self.host))
        methods = [
            responses.GET,
            responses.POST,
            responses.PUT,
            responses.DELETE,
            responses.PATCH,
        ]
        for method in methods:
            responses.add_callback(
                method, pattern,
                callback=_mock_callback
            )

    def get(self, *args, **kw):
        return self._do_request("get", *args, **kw)

    def put(self, *args, **kw):
        return self._do_request("put", *args, **kw)

    def post(self, *args, **kw):
        return self._do_request("post", *args, **kw)

    def delete(self, *args, **kw):
        return self._do_request("delete", *args, **kw)

    def patch(self, *args, **kw):
        return self._do_request("patch", *args, **kw)

    def setUp(self):
        pass

    def auth(self, username=None, player_name=None):
        """
        Do an auth call against the current app's /auth endpoint and fetch the
        root document which includes all user endpoints.
        """
        username = username or "systest"
        payload = {
            "provider": self.auth_provider,
            "username": username,
            "password": local_password,
        }
        resp = self.post("/auth", data=payload)

        self.token = resp.json()["token"]
        self.jti = resp.json()["jti"]
        self.current_user = jwt.decode(self.token, verify=False, algorithms=[JWT_ALGORITHM])
        self.player_id = self.current_user["player_id"]
        self.user_id = self.current_user["user_id"]
        self.headers = {"Authorization": "JWT " + self.token}
        self.player_name = player_name or username

        r = self.get("/")
        self.endpoints = r.json()["endpoints"]

    def auth_service(self):
        """
        Authenticate as a service user
        """
        payload = {
            "username": service_username,
            "password": service_password,
            "provider": self.auth_provider,
        }
        resp = self.post("/auth", data=payload)
        token = resp.json()["token"]
        jti = resp.json()["jti"]
        self.token = token
        self.jti = jti
        self.current_user = jwt.decode(self.token, verify=False, algorithms=[JWT_ALGORITHM])
        self.player_id = self.current_user["player_id"]
        self.user_id = self.current_user["user_id"]
        self.headers = {"Authorization": "JWT " + token, }
        self.player_name = service_username

        r = self.get("/")
        self.endpoints = r.json()["endpoints"]

    @classmethod
    def _authenticate_with_provider(cls, auth_info):
        user_name = auth_info['username'] or 'systest'
        user_id = abs(hash(user_name) & 0x7FFFFF)
        identity = {
            'user_name': auth_info['username'] or 'systest_user_name',
            'user_id': user_id,
            'identity_id': user_id + 1,
            'player_id': user_id + 2,
            'player_name': auth_info['username'] or 'systest_player_name',
            'roles': ['service'] if user_name == service_username else [],
        }
        return identity

    @classmethod
    def setUpClass(cls):
        ts = setup_tenant()
        cls.host = "http://localhost"
        app = drift_app()
        cls.app = app.test_client()

        # Add a 'unit_test' authentication provider
        register_auth_provider(app, AUTH_TEST_PROVIDER, cls._authenticate_with_provider)

        # HACK WARNING: drift-base implements its own authentication logic. If there is a
        # provider called 'user+pass' we will use that instead of the unit test version.
        if 'user+pass' in app.jwt_auth_providers:
            cls.auth_provider = 'user+pass'

    @classmethod
    def tearDownClass(cls):
        remove_tenant()



import unittest
import json
import pytest

from flask import request, jsonify
from flask.views import MethodView
from flask_smorest import Blueprint

from drift.tests import DriftTestCase
from drift.core.extensions.schemachecker import simple_schema_request, drift_init_extension


bp = Blueprint('schema', __name__, url_prefix='/schematest', description='test')

# we want the RuntimeError generaded by the response validator to be raised right through
# @api.errorhandler
# def handle(error):
#     raise type(error)(error.message)

@pytest.mark.skip(reason="This is quite broken and schemachecker is deprecated anyways.")
class SchemaCheckTest(DriftTestCase):

    def create_app(self):
        app = super(SchemaCheckTest, self).create_app()
        app.config['json_response_schema_validation'] = 'raise_exception'
        drift_init_extension(app, api=None)
        app.register_blueprint(bp)
        return app

    def test_required_property(self):
        # Test required property
        response = self.post(400, "/schematest", {})
        bla = json.loads(response.data.decode("ascii"))['description']
        self.assertIn("'string_required' is a required property", bla)

    def test_extra_unwanted(self):
        # Test extra unwanted property
        response = self.post(400, "/schematest", {"not_expected": 123})
        print("WHAT IS RESPONSE", response.data)
        self.assertEqual(response.status_code, 400, json.loads(response.data.decode('ascii'))['description'])
        self.assertIn("'additionalProperties': False", json.loads(response.data.decode('ascii')).get('description'))

    def test_successful_input(self):
        # Test successfull input data
        self.post(200, "/schematest", {"string_required": "x", "string_optional": "x"})

    def test_incorrect_response(self):
        # Test incorrect response
        with self.assertRaises(Exception) as context:
            self.post(400, "/schematest", {"string_required": "x", "fail_response": True})
            self.assertIn("'This is not expected!' is not of type 'integer'", str(context.exception))


@bp.route('', endpoint='health')
class SchemaTestAPI(MethodView):

    @simple_schema_request(
        {
            "string_required": {"type": "string", },
            "string_optional": {"type": "string", },
            "fail_response": {"type": "boolean", },
        },
        required=["string_required"]
    )
    def post(self):
        if request.json.get('fail_response'):
            return "This is not expected!"
        else:
            return jsonify({'response': 'fine'})


if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.CRITICAL)  # Quiet down logs
    unittest.main()



"""
Flask local development app.
Assigns a default tenant name 'developer'.
These setting can be overridden with env 'DRIFT_DEFAULT_TENANT'
"""
import logging
import os

from drift.flaskfactory import drift_app

# Default tenant for devapp is 'developer'
tenant_name = os.environ.setdefault('DRIFT_DEFAULT_TENANT', 'developer')

app = drift_app()

logging.getLogger(__name__).info("Default tenant for this developer app is '%s'.", tenant_name)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import os.path
import importlib
from contextlib import contextmanager
import socket
import getpass
import time

from six.moves import http_client

from click import echo

from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool

from werkzeug.local import LocalProxy
from sqlalchemy import create_engine
from flask import g, abort
from flask import _app_ctx_stack as stack, current_app

from flask_sqlalchemy import SQLAlchemy

from drift.flaskfactory import load_flask_config
from drift.core.extensions.driftconfig import check_tenant

import logging
log = logging.getLogger(__name__)


# defaults when making a new tier
TIER_DEFAULTS = {
    "server": "<PLEASE FILL IN>",
    "database": None,
    "port": 5432,
    "username": "zzp_user",
    "password": "zzp_user",
    "driver": "postgresql",
}


HAS_LOCAL_SERVER_MODE = True  # Supports DRIFT_USE_LOCAL_SERVERS flag.


def register_deployable(ts, deployablename, attributes):
    """
    Deployable registration callback.
    'deployablename' is from table 'deployable-names'.
    """
    # Make sure 'models' is specified in attributes
    if 'models' not in attributes:
        raise RuntimeError('''
'models' is a required attribute for postgres resource.
Add to config.json the name of the modules containing your SQLAlchemy db models.
Example:

    "resource_attributes": {
            "drift.core.resources.postgres": {
                "models": ["%s.db.models"]
            }
    }
    ''' % deployablename['deployable_name'].replace('-', ''))


def register_resource_on_tier(ts, tier, attributes):
    """
    Tier registration callback.
    'tier' is from table 'tiers'.
    'attributes' is a dict containing optional attributes for default values.
    """
    pass


def register_deployable_on_tier(ts, deployable, attributes):
    """
    Deployable registration callback.
    'deployable' is from table 'deployables'.
    """
    # Add default parameters for Postgres connection if needed.
    pass


def provision_resource(ts, tenant_config, attributes):
    """
    Create, recreate or delete resources for a tenant.
    'tenant_config' is a row from 'tenants' table for the particular tenant, tier and deployable.
    LEGACY SUPPORT: 'attributes' points to the current resource attributes within 'tenant_config'.
    """

    if tenant_config['state'] == 'deleted':
        return ["DB is, or should be deleted. Not bothering to check though."]
        return

    # Create the tier default user on the DBMS and assign "can login" privilege
    # and add role "rds_superuser".
    params = process_connection_values(attributes)
    params["username"] = MASTER_USER
    params["password"] = MASTER_PASSWORD
    params["database"] = MASTER_DB
    engine = connect(params)
    params['role'] = 'rds_superuser'
    sql = "CREATE ROLE zzp_user LOGIN PASSWORD 'zzp_user' VALID UNTIL 'infinity';"
    try:
        engine.execute(sql)
    except Exception as e:
        if "already exists" not in str(e):
            raise

    sql = "GRANT {role} TO zzp_user;".format(**params)
    try:
        engine.execute(sql)
    except Exception as e:
        if "role \"{role}\" does not exist".format(**params) not in str(e):
            raise

    report = []

    # LEGACY SUPPORT:
    # The old provision logic would load in config.json on its own to get the app name. Now we simply
    # inject this info into the attributes/params dict.
    attributes['application_name'] = tenant_config['deployable_name']

    # MORE LEGACY STUFF:
    # The db model module names are registered in 'deployable-names' table. We inject this info into
    # the attributes/params dict.
    depl = ts.get_table('deployable-names').get({'deployable_name': tenant_config['deployable_name']})
    attributes['models'] = depl['resource_attributes']['drift.core.resources.postgres']['models']

    # Initialize the DB name if applicable
    if not attributes.get('database'):
        attributes["database"] = "{}.{}".format(
            tenant_config['tenant_name'], tenant_config['deployable_name'])

    attributes = attributes.copy()  # Do not give subroutines chance of modifying permanently.

    if os.environ.get('DRIFT_USE_LOCAL_SERVERS', False):
        # Override 'server'
        attributes['server'] = os.environ.get('DRIFT_POSTGRES_HOST', 'localhost')

    if tenant_config['state'] == 'initializing':
        # Create or recreate db
        if db_exists(attributes):
            drop_db(attributes, force=True)
            report.append("Database for tenant already existed, dropped the old DB.")

        create_db(attributes, report=report)
    elif tenant_config['state'] == 'active':
        if not db_exists(attributes):
            log.warning(
                "Database for tenant '%s' doesn't exist, which is unexpected. Creating one now.",
                tenant_config['tenant_name']
            )
            report.append("Database didn't exist, which was unexpected, so a new DB needs to be created")
            create_db(attributes, report=report)
        report.append("Database check successful for DB: {}".format(format_connection_string(attributes)))
    elif tenant_config['state'] == 'uninitializing':
        # Archive or delete db
        drop_db(attributes, force=True)

    return report


# we need a single master db on all db instances to perform db maintenance
MASTER_DB = 'postgres'
MASTER_USER = 'postgres'
MASTER_PASSWORD = 'postgres'
ECHO_SQL = False
SCHEMAS = ["public"]


class Postgres(object):
    """Postgres Flask extension."""

    application_name = None  # Set just in time to something sensible (see below)

    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if not hasattr(app, 'extensions'):
            app.extensions = {}

        app.extensions['postgres'] = self
        app.before_request(self.before_request)
        app.teardown_request(self.teardown_request)

    def before_request(self):
        # Add a just-in-time getter for session
        g.db = LocalProxy(self.get_session)

    def teardown_request(self, exception):
        """Return the database connection at the end of the request"""
        ctx = stack.top
        if ctx is not None:
            if hasattr(ctx, 'sqlalchemy_session'):
                try:
                    ctx.sqlalchemy_session.close()
                except Exception as e:
                    log.error("Could not close sqlalchemy session: %s", e)

    def get_session(self):
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'sqlalchemy_session'):
                ctx.sqlalchemy_session = get_sqlalchemy_session()
            return ctx.sqlalchemy_session

    @classmethod
    def get_application_name(cls):

        if cls.application_name is None:
            # Pretty print a nice name for this connection, just in time.
            cls.application_name = '{}:{}@{}'.format(
                load_flask_config()['name'],
                getpass.getuser(),
                socket.gethostname().split(".")[0],
            )

        return cls.application_name


def drift_init_extension(app, **kwargs):
    Postgres(app)

    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///:memory:'  # Just to quiet down a warning
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False  # https://tinyurl.com/SQLALCHEMY-TRACK-MODIFICATIONS
    app.config.setdefault('SQLALCHEMY_DATABASE_OPTIONS', {}).update({
        'connect_args': {
            'connect_timeout': 10,
            'application_name': Postgres.get_application_name(),
        }
    })
    SQLAlchemy(app)


@check_tenant
def get_sqlalchemy_session(conn_string=None):
    """
    Return an SQLAlchemy session for the specified DB connection string
    """
    if not conn_string:
        if not g.conf.tenant:
            abort(http_client.BAD_REQUEST, "No DB resource available because no tenant is specified.")

        ci = g.conf.tenant.get('postgres')
        conn_string = format_connection_string(ci)

    log.debug("Creating sqlalchemy session with connection string '%s'", conn_string)

    # Flaschemy injection
    # Populate binds just-in-time. The middleware takes care of rest
    current_app.config['SQLALCHEMY_BINDS'] = {g.conf.tenant['tenant_name']: conn_string}
    engine = current_app.extensions['sqlalchemy'].db.get_engine(bind=g.conf.tenant['tenant_name'])
    session = current_app.extensions['sqlalchemy'].db.create_scoped_session(options={'bind': engine})
    return session


@contextmanager
def sqlalchemy_session(conn_string=None):
    session = get_sqlalchemy_session(conn_string)
    try:
        yield session
        session.commit()
    except BaseException:
        session.rollback()
        raise
    finally:
        session.close()


def process_connection_values(postgres_parameters):
    """
    Returns a copy of 'postgres_parameters' where values have been
    processed or or overridden if applicable.
    """
    postgres_parameters = postgres_parameters.copy()
    if os.environ.get('DRIFT_USE_LOCAL_SERVERS', False):
        # Override 'server'
        postgres_parameters['server'] = os.environ.get('DRIFT_POSTGRES_HOST', 'localhost')
    return postgres_parameters


def format_connection_string(postgres_parameters):
    postgres_parameters = process_connection_values(postgres_parameters)
    connection_string = '{driver}://{username}:{password}@{server}:{port}/{database}'.format(**postgres_parameters)
    return connection_string


def connect(params, connect_timeout=None):

    engine = create_engine(
        format_connection_string(params),
        echo=ECHO_SQL,
        isolation_level='AUTOCOMMIT',
        connect_args={
            'connect_timeout': connect_timeout or 10,
            'application_name': params.get('application_name', 'drift.core.resources.postgres'),
        }
    )
    return engine


def db_exists(params):
    try:
        engine = connect(params)
        engine.execute("SELECT 1=1")
    except Exception as e:
        if "does not exist" in repr(e):
            return False
        else:
            echo("OOPS: {!r}".format(e))
            return False
    return True


def db_check(params):
    """
    Do a simple check on DB referenced by 'params' and return a string describing any error
    that may occur. If all is fine, None is returned.
    """
    try:
        engine = connect(params, connect_timeout=2)
        engine.execute("SELECT 1=1")
    except Exception as e:
        return str(e)


def create_db(params, report=None):

    # These imports have significant performance impact so that's why they are not at file scope.
    from alembic.config import Config
    from alembic import command

    t = time.time()  # Simple timing of CB creation.
    params = process_connection_values(params)
    db_name = params["database"]
    username = params["username"]

    params["username"] = MASTER_USER
    params["password"] = MASTER_PASSWORD

    master_params = params.copy()
    master_params["database"] = MASTER_DB
    engine = connect(master_params)
    engine.execute('COMMIT')
    sql = 'CREATE DATABASE "{}";'.format(db_name)
    try:
        engine.execute(sql)
    except Exception as e:
        echo("{!r} {!r}".format(sql, e))

    # TODO: This will only run for the first time and fail in all other cases.
    # Maybe test before instead?
    sql = "CREATE ROLE zzp_user LOGIN PASSWORD 'zzp_user' VALID UNTIL 'infinity';"
    try:
        engine.execute(sql)
    except Exception as e:
        pass

    engine = connect(params)

    # TODO: Alembic (and sqlalchemy for that matter) don't like schemas. We should
    # figure out a way to add these later
    # LEGACY SUPPORT: We can't load config.json arbitrarily here so instead we just assume
    # db model modules are found in "<flugger>"
    models = params.get("models", [])
    if not models:
        raise Exception("This app has no models defined in config")

    for model_module_name in models:
        log.info("Building models from %s", model_module_name)
        models = importlib.import_module(model_module_name)
        if hasattr(models, 'pre_create_db_tables'):
            models.pre_create_db_tables(engine)
        models.ModelBase.metadata.create_all(engine)
        if hasattr(models, 'on_create_db'):
            models.on_create_db(engine)

    # stamp the db with the latest alembic upgrade version
    alembic_not_supported = True
    if alembic_not_supported:
        log.warning("NOTE! Alembic not supported at the moment. No db upgrades are run.")
    else:
        from drift.utils import get_app_root
        approot = get_app_root()
        ini_path = os.path.join(approot, "alembic.ini")
        alembic_cfg = Config(ini_path)
        script_path = os.path.join(os.path.split(os.path.abspath(ini_path))[0], "alembic")
        alembic_cfg.set_main_option("script_location", script_path)
        db_names = alembic_cfg.get_main_option('databases')
        connection_string = format_connection_string(params)
        alembic_cfg.set_section_option(db_names, "sqlalchemy.url", connection_string)
        command.stamp(alembic_cfg, "head")

    for schema in SCHEMAS:
        # Note that this does not automatically grant on tables added later
        sql = '''
                 GRANT SELECT, INSERT, UPDATE, DELETE ON ALL TABLES IN SCHEMA "{schema}" TO {user};
                 GRANT USAGE, SELECT, UPDATE ON ALL SEQUENCES IN SCHEMA "{schema}" TO {user};
                 GRANT ALL ON SCHEMA "{schema}" TO {user};'''.format(schema=schema, user=username)
        try:
            engine.execute(sql)
        except Exception as e:
            echo("{!r} {!r}".format(sql, e))

    if report:
        report.append("Created a new DB: '{}' in {:.3f} seconds.".format(db_name, time.time() - t))

    return db_name


def drop_db(_params, force=False):
    params = process_connection_values(_params)
    db_name = params["database"]
    if 'test' not in db_name.lower() and not force:
        raise RuntimeError("Will not drop database '{}' because it's not a test DB. Use 'force' to override".format(db_name))

    params["database"] = MASTER_DB
    params["username"] = MASTER_USER
    params["password"] = MASTER_PASSWORD

    if not db_exists(params):
        echo("Not dropping database {!r} as it doesn't seem to exist.".format(db_name))
        return

    engine = connect(params)

    # disconnect connected clients
    engine.execute("SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname = '{}';"
                   .format(db_name))

    sql = 'DROP DATABASE "{}";'.format(db_name)
    engine.execute('COMMIT')

    try:
        engine.execute(sql)
    except Exception as e:
        echo("{!r} {!r}".format(sql, e))

    log.info("Database '%s' has been dropped on '%s'", db_name, params["server"])


def healthcheck():
    if "postgres" not in g.conf.tenant:
        abort(http_client.SERVICE_UNAVAILABLE, "Tenant config does not have 'postgres' section.")
    for k in TIER_DEFAULTS.keys():
        if not g.conf.tenant["postgres"].get(k):
            abort(http_client.SERVICE_UNAVAILABLE, "'postgres' config missing key '%s'" % k)

    rows = g.db.execute("SELECT 1+1")
    rows.fetchall()[0]



import warnings

from flask import current_app


# This goes away once we stop using the global @register_endpoints decorator
needs_to_be_global = []


class Endpoints(object):
    """
    A class to register endpoint defitions functions
    at import time
    """
    def __init__(self):
        self.registry_funcs = []

    def init_app(self, app):
        """
        when app is initialized, the registered endpoints are handed to the registry
        """
        for f in self.registry_funcs:
            registry.register_app_endpoints(app, f)

    def register(self, f):
        """
        At import time, register endpoint functions here
        """
        self.registry_funcs.append(f)
        return f


class EndpointRegistry(object):

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        if 'urlregistry' not in app.extensions:
            app.extensions['urlregistry'] = self
        if not hasattr(app, "endpoint_registry_funcs"):
            app.endpoint_registry_funcs = needs_to_be_global  # used while we still have static registration
            app.endpoint_registry_funcs2 = []  # proper per-app registry

    def register_app_endpoints(self, app, f):
        self.init_app(app)
        app.endpoint_registry_funcs2.append(f)

    # legacy function, going away. used by static assignation and modifies the static method.
    def register_endpoints(self, f):
        app = current_app
        self.init_app(app)
        app.endpoint_registry_funcs.append(f)
        return f


def register_endpoints(f):
    warnings.warn(
        "please use Endpoints().register instead of register_endpoints",
        DeprecationWarning,
        stacklevel=2)
    # TODO: It is a bad pattern to import app here since it might cause the app to be created
    registry.register_endpoints(f)
    return f


def drift_init_extension(app, **kwargs):
    registry.init_app(app)


# the static registry object.  Contains no data, just a plain flask extension
registry = EndpointRegistry()



"""
Build an AWS AMI for this service
"""
import os
import sys
import time
import subprocess
import operator
import pkg_resources
import json
from datetime import datetime, timedelta
import pytz
import random
import shlex
import tempfile

from click import echo, secho
import six
from six import print_

try:
    # boto library is not a hard requirement for drift.
    import boto3
except ImportError:
    pass

from drift.management import get_app_version, create_deployment_manifest
from drift.management.gittools import get_branch, checkout
from drift.utils import get_tier_name
from driftconfig.util import get_drift_config
from driftconfig.config import get_redis_cache_backend
from drift.flaskfactory import load_flask_config


UBUNTU_TRUSTY_IMAGE_NAME = 'ubuntu/images/hvm/ubuntu-trusty-14.04*'
UBUNTU_XENIAL_IMAGE_NAME = 'ubuntu/images/hvm-ssd/ubuntu-xenial-16.04*'
UBUNTU_BIONIC_IMAGE_NAME = 'ubuntu/images/hvm-ssd/ubuntu-bionic-18.04*'
UBUNTU_RELEASE = UBUNTU_BIONIC_IMAGE_NAME

IAM_ROLE = "ec2"


# The 'Canonical' owner. This organization maintains the Ubuntu AMI's on AWS.
def _get_owner_id_for_canonical(region_id):
    """Returns region specific owner id for Canonical which is the maintainer of Ubuntu images."""
    if region_id.startswith('cn-'):
        return '837727238323'
    else:
        return '099720109477'


def get_options(parser):

    subparsers = parser.add_subparsers(
        title="AWS AMI Management and Deployment",
        description="These sets of commands help you with configuring, baking, running and "
        "deploying AWS AMIs' for your tier.",
        dest="command",
    )
    subparsers.required = True

    # The 'bake' command
    p = subparsers.add_parser(
        'bake',
        help='Bake a new AMI for the current current service.',
    )
    p.add_argument(
        'tag', action='store', help='Git release tag to bake. (Run "git tag" to get available tags).',
                                    nargs='?', default=None)
    p.add_argument(
        "--preview", help="Show arguments only", action="store_true"
    )
    p.add_argument(
        "--skipcopy", help="Do not copy image to all regions", action="store_true"
    )
    p.add_argument(
        "--debug", help="Run Packer in debug mode", action="store_true"
    )
    p.add_argument(
        "--ami",
        help="Specify base image. Default is the latest Ubuntu image from Canonical."
    )
    p.add_argument(
        "--instance_type",
        help="The EC2 instance type to use to build. Default is 'm4.xlarge'.",
        default="m4.xlarge"
    )

    # The 'run' command
    p = subparsers.add_parser(
        'run',
        help='Launch an AMI for this service, or configure it for auto-scaling.',
    )
    p.add_argument(
        "--ami",
        help="An AMI built with the rest api service",
    )
    p.add_argument(
        "--instance_type",
        help="The EC2 instance type to use. Default is 't2.small'.",
        default="t2.small"
    )
    p.add_argument(
        "--launch", action="store_true",
        help="Launch the AMI. (Default unless \"autoscaling\" is configured for the tier and service.)",
    )
    p.add_argument(
        "--autoscale", action="store_true",
        help="Add the AMI to autoscaling group. (Default if \"autoscaling\" is configured for the tier and service.)",
    )
    p.add_argument(
        "--force", action="store_true",
        help="If --launch or --autoscale conflict with tier configuration, use --force to override.",
    )
    p.add_argument(
        "--preview", help="Show arguments only", action="store_true"
    )
    p.add_argument(
        "--verbose", help="Verbose output", action="store_true"
    )

    # The 'copy-image' command
    p = subparsers.add_parser(
        'copy-image',
        help='Copies AMI to all active regions.',
    )
    p.add_argument(
        "ami",
        help="The image id.",
    )


def run_command(args):
    fn = globals()["_{}_command".format(args.command.replace("-", "_"))]
    fn(args)


def fold_tags(tags, key_name=None, value_name=None):
    """Fold boto3 resource tags array into a dictionary."""
    return {tag['Key']: tag['Value'] for tag in tags}


def filterize(d):
    """
    Return dictionary 'd' as a boto3 "filters" object by unfolding it to a list of
    dict with 'Name' and 'Values' entries.
    """
    return [{'Name': k, 'Values': [v]} for k, v in d.items()]


def _bake_command(args):
    conf = get_drift_config(drift_app=load_flask_config())
    name = conf.drift_app['name']

    domain = conf.domain.get()
    if 'aws' not in domain or 'ami_baking_region' not in domain['aws']:
        echo(
            "Missing configuration value in table 'domain'. Specify the AWS region in "
            "'aws.ami_baking_region'.")
        sys.exit(1)
    aws_region = domain['aws']['ami_baking_region']
    ec2 = boto3.resource('ec2', region_name=aws_region)

    # Do some compatibility checks
    if aws_region.startswith("cn-"):
        echo("NOTE!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        echo("If this command fails with some AWS access crap, you may need to switch packer version.")
        echo("More info: https://github.com/hashicorp/packer/issues/5447")

    echo("DOMAIN:")
    echo(json.dumps(domain, indent=4))
    echo("DEPLOYABLE: {!r}".format(name))
    echo("AWS REGION: {!r}".format(aws_region))

    # Clean up lingering Packer instances
    tag_name = 'Packer Builder'
    ec2_client = boto3.client('ec2', region_name=aws_region)
    packers = ec2_client.describe_instances(
        Filters=[
            {'Name': 'tag:Name', 'Values': ['Packer Builder']},
            {'Name': 'instance-state-name', 'Values': ['running']}
        ]
    )
    terminate_ids = []
    if packers['Reservations']:
        for pc2 in packers['Reservations'][0]['Instances']:
            age = datetime.utcnow().replace(tzinfo=pytz.utc) - pc2['LaunchTime']
            if age > timedelta(minutes=20):
                echo("Cleaning up Packer instance {} as it has been active for {}.".format(
                    pc2['InstanceId'], age))
                terminate_ids.append(pc2['InstanceId'])
        if terminate_ids:
            ec2.instances.filter(InstanceIds=terminate_ids).terminate()

    if args.ami:
        amis = list(ec2.images.filter(ImageIds=[args.ami]))
        ami = amis[0]
    else:
        # Get all Ubuntu images from the appropriate region and pick the most recent one.
        # The 'Canonical' owner. This organization maintains the Ubuntu AMI's on AWS.
        echo("Finding the latest AMI on AWS that matches {!r}".format(UBUNTU_RELEASE))
        filters = [
            {'Name': 'name', 'Values': [UBUNTU_RELEASE]},
            {'Name': 'architecture', 'Values': ['x86_64']},
        ]
        amis = list(ec2.images.filter(Owners=[_get_owner_id_for_canonical(aws_region)], Filters=filters))
        if not amis:
            echo("No AMI found matching {!r}. Not sure what to do now.".format(UBUNTU_RELEASE))
            sys.exit(1)
        ami = max(amis, key=operator.attrgetter("creation_date"))

    echo("Using source AMI:")
    echo("\tID:\t{!r}".format(ami.id))
    echo("\tName:\t{!r}".format(ami.name))
    echo("\tDate:\t{!r}".format(ami.creation_date))

    current_branch = get_branch()
    if not args.tag:
        args.tag = current_branch

    echo("Using branch/tag {!r}".format(args.tag))

    # Wrap git branch modification in RAII.
    checkout(args.tag)
    try:
        setup_script = ""
        setup_script_custom = ""
        with open(pkg_resources.resource_filename(__name__, "driftapp-packer.sh"), 'r') as f:
            setup_script = f.read()
        custom_script_name = os.path.join(conf.drift_app['app_root'], 'scripts', 'ami-bake.sh')
        if os.path.exists(custom_script_name):
            echo("Using custom bake shell script {!r}".format(custom_script_name))
            setup_script_custom = "echo Executing custom bake shell script from {}\n".format(custom_script_name)
            setup_script_custom += open(custom_script_name, 'r').read()
            setup_script_custom += "\necho Custom bake shell script completed\n"
        else:
            echo("Note: No custom ami-bake.sh script found for this application.")
        # custom setup needs to happen first because we might be installing some requirements for the regular setup
        setup_script = setup_script_custom + setup_script
        with tempfile.NamedTemporaryFile('w', delete=False) as tf:
            tf.write(setup_script)

        setup_script_filename = tf.name
        manifest = create_deployment_manifest('ami', comment=None, deployable_name=name)
        packer_vars = {
            'version': get_app_version(),
            'setup_script': setup_script_filename,
        }

        if not args.preview:
            # TODO: This code mirrors the one in ami.py. It's not DRY.
            cmd = [sys.executable, 'setup.py', 'sdist', '--formats=tar']
            ret = subprocess.call(cmd)
            if ret != 0:
                secho("Failed to execute build command: {!r}".format(cmd), fg="red")
                sys.exit(ret)

            import shutil
            shutil.make_archive("dist/aws", 'tar', "aws")
    finally:
        echo("Reverting to {!r}".format(current_branch))
        checkout(current_branch)

    packer_vars.update({
        "instance_type": args.instance_type,
        "service": name,
        "region": aws_region,
        "source_ami": ami.id,
        "user_name": boto3.client('sts').get_caller_identity()['Arn'],
        "domain_name": domain['domain_name'],
    })

    echo("Packer variables:")
    echo(pretty(packer_vars))

    # See if Packer is installed and generate sensible error code if something is off.
    # This will also write the Packer version to the terminal which is useful info.
    try:
        subprocess.call(['packer', 'version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except Exception as e:
        echo("Error: {}".format(e))
        echo("'packer version' command failed. Make sure it's installed.")
        if sys.platform == 'win32':
            echo("To install packer for windows: choco install packer")
        sys.exit(127)

    cmd = "packer build "
    if args.debug:
        cmd += "-debug "

    cmd += "-only=amazon-ebs "
    for k, v in packer_vars.items():
        cmd += "-var {}=\"{}\" ".format(k, v)

    cmd = shlex.split(cmd)

    # Use generic packer script if project doesn't specify one
    pkg_resources.cleanup_resources()
    if os.path.exists("config/packer.json"):
        cmd.append("config/packer.json")
    else:
        scriptfile = pkg_resources.resource_filename(__name__, "driftapp-packer.json")
        cmd.append(scriptfile)

    echo("Baking AMI with: {}".format(' '.join(cmd)))
    if args.preview:
        echo("Manifest tags assigned to AMI:")
        echo(pretty(manifest))

        echo("Not building or packaging because --preview is on. Exiting now.")
        sys.exit(0)

    start_time = time.time()
    failure_line = None

    try:
        # Execute Packer command and parse the output to find the ami id.
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        while True:
            line = p.stdout.readline()
            # packer is streaming stuff from the remote which uses utf-8 encoding.
            # in py2, we just leave the line as it is, gobble it and print it.
            if six.PY3:
                line = line.decode()
            print_(line, end="")
            if line == '' and p.poll() is not None:
                break

            # The last lines from the packer execution look like this:
            # ==> Builds finished. The artifacts of successful builds are:
            # --> amazon-ebs: AMIs were created:
            #
            # eu-west-1: ami-0ee5eb68
            if 'ami-' in line:
                ami_id = line[line.rfind('ami-'):].strip()
                ami = ec2.Image(ami_id)
                echo()
                echo("AMI ID: %s" % ami.id)
                echo()

            if "failed with error code" in line:
                failure_line = line
            if "Your Pipfile.lock" in line and "is out of date" in line:
                secho("ERROR: {}".format(line), fg="red")
                secho("Build failed! Consider running 'pipenv lock' before baking.", fg="red")
                sys.exit(1)
            if "Creating a Pipfile for this project" in line:
                secho("ERROR: No Pipfile in distribution! Check 'MANIFEST.in'.", fg="red")
                sys.exit(1)
            if "Pipfile.lock not found, creating" in line:
                secho("ERROR: No Pipfile.lock in distribution! Check 'MANIFEST.in'.", fg="red")
                sys.exit(1)
    finally:
        pkg_resources.cleanup_resources()

    if p.returncode != 0:
        secho("Failed to execute packer command: {!r}".format(cmd), fg="red")
        if failure_line:
            secho("Check this out: {!r}".format(failure_line), fg="yellow")
        sys.exit(p.returncode)

    duration = time.time() - start_time

    if manifest:
        echo("Adding manifest tags to AMI:")
        echo(pretty(manifest))
        prefix = "drift:manifest:"
        tags = []
        for k, v in manifest.items():
            tag_name = "{}{}".format(prefix, k)
            tags.append({'Key': tag_name, 'Value': v or ''})
        ami.create_tags(DryRun=False, Tags=tags)

    if not args.skipcopy:
        _copy_image(ami.id)

    secho("Done after %.0f seconds" % (duration,), fg="green")


class MyEncoder(json.JSONEncoder):
    def default(self, o):
        return str(o)


def pretty(ob):
    """Returns a pretty representation of 'ob'."""
    return json.dumps(ob, cls=MyEncoder, indent=4)


def _find_latest_ami(service_name, release=None):
    tier_name = get_tier_name()
    conf = get_drift_config(tier_name=tier_name, deployable_name=service_name)
    domain = conf.domain.get()
    aws_region = conf.tier['aws']['region']

    ec2 = boto3.resource('ec2', region_name=aws_region)
    filters = [
        {'Name': 'tag:service-name', 'Values': [service_name]},
        {'Name': 'tag:domain-name', 'Values': [domain['domain_name']]},
    ]
    if release:
        filters.append({'Name': 'tag:git-release', 'Values': [release]},)

    amis = list(ec2.images.filter(Owners=['self'], Filters=filters))
    if not amis:
        criteria = {d['Name']: d['Values'][0] for d in filters}
        secho("No AMI found using the search criteria {}.".format(criteria), fg="red")
        sys.exit(1)

    ami = max(amis, key=operator.attrgetter("creation_date"))
    return ami


def _run_command(args):
    # Always autoscale!
    args.autoscale = True

    if args.launch and args.autoscale:
        secho("Error: Can't use --launch and --autoscale together.", fg="red")
        sys.exit(1)

    tier_name = get_tier_name()
    conf = get_drift_config(tier_name=tier_name, drift_app=load_flask_config())
    name = conf.drift_app['name']

    if not conf.deployable:
        echo("The deployable '{}' is not registered and/or assigned to tier {!r}.".format(name, tier_name))
        echo("Run 'drift-admin register' to register this deployable.")
        echo("Run 'driftconfig assign-tier {}' to assign it to the tier.".format(name))
        sys.exit(1)

    aws_region = conf.tier['aws']['region']

    if args.verbose:
        echo("AWS REGION: {!r}".format(aws_region))
        echo("DOMAIN:")
        echo(json.dumps(conf.domain.get(), indent=4))
        echo("DEPLOYABLE:")
        echo(json.dumps(conf.deployable, indent=4))

    ec2_conn = boto3.resource('ec2', region_name=aws_region)

    if conf.tier['is_live']:
        secho("NOTE! This tier is marked as LIVE. Special restrictions may apply. Use --force to override.", fg="yellow")

    autoscaling = {
        "min": 1,
        "max": 1,
        "desired": 1,
        "instance_type": args.instance_type,
    }
    autoscaling.update(conf.deployable.get('autoscaling', {}))
    release = conf.deployable.get('release', '')

    if args.launch and autoscaling and not args.force:
        secho("--launch specified, but tier config specifies 'use_autoscaling'. Use --force to ovefrride.", fg="red")
        sys.exit(1)
    if args.autoscale and not autoscaling and not args.force:
        secho("--autoscale specified, but tier config doesn't specify 'use_autoscaling'. Use --force to override.", fg="red")
        sys.exit(1)

    echo("Launch an instance of {!r} on tier {!r}".format(name, tier_name))
    if release:
        echo("Using AMI with release tag: {!r}".format(release))
    else:
        echo("Using the newest AMI baked (which may not be what you expect).")

    ami = _find_latest_ami(name, release)
    echo("AMI: {} [{}]".format(ami.id, ami.name))

    if args.ami:
        echo("Using a specified AMI: {!r}".format(args.ami))
        ec2 = boto3.resource('ec2', region_name=aws_region)
        if ami.id != args.ami:
            secho("AMI found is different from AMI specified on command line.", fg="yellow")
            if conf.tier['is_live'] and not args.force:
                secho("This is a live tier. Can't run mismatched AMI unless --force is specified", fg="red")
                sys.exit(1)
        try:
            ami = ec2.Image(args.ami)
        except Exception as e:
            raise RuntimeError("Ami '%s' not found or broken: %s" % (args.ami, e))

    if not ami:
        sys.exit(1)

    ami_info = dict(
        ami_id=ami.id,
        ami_name=ami.name,
        ami_created=ami.creation_date,
        ami_tags={d['Key']: d['Value'] for d in ami.tags},
    )

    if args.verbose:
        echo("AMI Info:")
        echo(pretty(ami_info))

        if autoscaling:
            echo("Autoscaling group:")
            echo(pretty(autoscaling))
        else:
            echo("EC2:")
            echo("\tInstance Type:\t{}".format(args.instance_type))

    ec2 = boto3.resource('ec2', region_name=aws_region)

    # Get all 'private' subnets
    filters = {'tag:tier': tier_name, 'tag:realm': 'private'}
    subnets = list(ec2.subnets.filter(Filters=filterize(filters)))
    if not subnets:
        secho("Error: No subnet available matching filter {}".format(filters), fg="red")
        sys.exit(1)

    # Get the "one size fits all" security group
    filters = {'tag:tier': tier_name, 'tag:Name': '{}-private-sg'.format(tier_name)}
    security_group = list(ec2.security_groups.filter(Filters=filterize(filters)))[0]

    # The key pair name for SSH
    key_name = conf.tier['aws']['ssh_key']
    if "." in key_name:
        key_name = key_name.split(".", 1)[0]  # TODO: Distinguish between key name and .pem key file name

    '''
    autoscaling group:
    Name            LIVENORTH-themachines-backend-auto
    api-port        10080
    api-target      themachines-backend
    service-name    themachines-backend
    service-type    rest-api
    tier            LIVENORTH

    ec2:
    Name            DEVNORTH-drift-base
    launched-by     nonnib
    api-port        10080
    api-target      drift-base
    service-name    drift-base
    service-type    rest-api
    tier            DEVNORTH
    '''

    target_name = "{}-{}".format(tier_name, name)
    if autoscaling:
        target_name += "-auto"

    # To auto-generate Redis cache url, we create the Redis backend using our config,
    # and then ask for a url representation of it:
    drift_config_url = get_redis_cache_backend(conf.table_store, tier_name).get_url()

    # Specify the app
    app_root = '/etc/opt/{service_name}'.format(service_name=name)

    tags = {
        "Name": target_name,
        "tier": tier_name,
        "service-name": name,
        "service-type": conf.drift_app.get('service_type', 'web-app'),
        "config-url": drift_config_url,
        "app-root": app_root,
        "launched-by": boto3.client('sts').get_caller_identity()['Arn'],
    }

    if tags['service-type'] == 'web-app':
        # Make instance part of api-router round-robin load balancing
        tags.update(
            {
                "api-target": name,
                "api-port": str(conf.drift_app.get('PORT', 10080)),
                "api-status": "online",
            }
        )

    tags.update(fold_tags(ami.tags))

    user_data = '''#!/bin/bash
# Environment variables set by drift-admin run command:
export DRIFT_CONFIG_URL={drift_config_url}
export DRIFT_TIER={tier_name}
export DRIFT_APP_ROOT={app_root}
export DRIFT_SERVICE={service_name}
export AWS_REGION={aws_region}

# Shell script from ami-run.sh:
'''.format(drift_config_url=drift_config_url, tier_name=tier_name, app_root=app_root, service_name=name, aws_region=aws_region)

    user_data += pkg_resources.resource_string(__name__, "ami-run.sh").decode()
    custom_script_name = os.path.join(conf.drift_app['app_root'], 'scripts', 'ami-run.sh')
    if os.path.exists(custom_script_name):
        echo("Using custom shell script {!r}".format(custom_script_name))
        user_data += "\n# Custom shell script from {}\n".format(custom_script_name)
        user_data += open(custom_script_name, 'r').read()
    else:
        echo("Note: No custom ami-run.sh script found for this application.")

    if args.verbose:
        echo("Subnets:")
        for subnet in subnets:
            echo("\t{} - {}".format(fold_tags(subnet.tags)['Name'], subnet.id))

        echo("Security Group:\n\t{} [{} {}]".format(fold_tags(security_group.tags)["Name"], security_group.id, security_group.vpc_id))

        echo("SSH Key:")
        echo(key_name)

        echo("Tags:")
        for k in sorted(tags.keys()):
            echo("  %s: %s" % (k, tags[k]))

        echo("user_data:")
        from drift.utils import pretty as poo
        echo(poo(user_data, 'bash'))

    if args.preview:
        echo("--preview specified, exiting now before actually doing anything.")
        sys.exit(0)

    user_data = user_data.replace('\r\n', '\n')

    if autoscaling:
        client = boto3.client('autoscaling', region_name=aws_region)
        launch_config_name = '{}-{}-launchconfig-{}-{}'.format(tier_name, name, datetime.utcnow(), release)
        launch_config_name = launch_config_name.replace(':', '.')

        kwargs = dict(
            LaunchConfigurationName=launch_config_name,
            ImageId=ami.id,
            KeyName=key_name,
            SecurityGroups=[security_group.id],
            InstanceType=args.instance_type or autoscaling['instance_type'],
            IamInstanceProfile=IAM_ROLE,
            InstanceMonitoring={'Enabled': True},
            UserData=user_data,
        )

        if args.verbose:
            echo("Creating launch configuration using params:")
            echo(pretty(kwargs))
        else:
            echo("Creating launch configuration: {}".format(launch_config_name))

        client.create_launch_configuration(**kwargs)

        # Update current autoscaling group or create a new one if it doesn't exist.
        groups = client.describe_auto_scaling_groups(AutoScalingGroupNames=[target_name])

        kwargs = dict(
            AutoScalingGroupName=target_name,
            LaunchConfigurationName=launch_config_name,
            MinSize=autoscaling['min'],
            MaxSize=autoscaling['max'],
            DesiredCapacity=autoscaling['desired'],
            VPCZoneIdentifier=','.join([subnet.id for subnet in subnets]),
        )

        if not groups['AutoScalingGroups']:
            echo("Creating a new autoscaling group using params:")
            echo(pretty(kwargs))
            client.create_auto_scaling_group(**kwargs)
        else:
            echo("Updating current autoscaling group {!r}".format(target_name))
            client.update_auto_scaling_group(**kwargs)

        # Prepare tags which get propagated to all new instances
        tagsarg = [
            {
                'ResourceId': tags['Name'],
                'ResourceType': 'auto-scaling-group',
                'Key': k,
                'Value': v,
                'PropagateAtLaunch': True,
            }
            for k, v in tags.items()
        ]
        echo("Updating tags on autoscaling group that get propagated to all new instances.")
        client.create_or_update_tags(Tags=tagsarg)

        # Define a 2 min termination cooldown so api-router can drain the connections.
        echo("Configuring lifecycle hook.")
        response = client.put_lifecycle_hook(
            LifecycleHookName='Wait-2-minutes-on-termination',
            AutoScalingGroupName=target_name,
            LifecycleTransition='autoscaling:EC2_INSTANCE_TERMINATING',
            HeartbeatTimeout=120,
            DefaultResult='CONTINUE'
        )

        echo("Terminating instances in autoscaling group. New ones will be launched.")
        echo("Old instances will linger for 2 minutes while connections are drained.")
        asg = client.describe_auto_scaling_groups(AutoScalingGroupNames=[target_name])
        for instance in asg['AutoScalingGroups'][0]['Instances']:
            response = client.terminate_instance_in_auto_scaling_group(
                InstanceId=instance['InstanceId'],
                ShouldDecrementDesiredCapacity=False
            )
            echo("   " + response['Activity']['Description'])

        secho("Done!", fg="green")

    else:
        # Pick a random subnet from list of available subnets
        subnet = random.choice(subnets)
        echo("Randomly picked this subnet to use: {!r}".format(subnet))

        echo("Launching EC2 instance...")
        reservation = ec2_conn.run_instances(
            ami.id,
            instance_type=args.instance_type,
            subnet_id=subnet.id,
            security_group_ids=[security_group.id],
            key_name=key_name,
            instance_profile_name=IAM_ROLE,
            user_data=user_data,
        )

        if len(reservation.instances) == 0:
            secho("No instances in reservation!", fg="red")
            sys.exit(1)

        instance = reservation.instances[0]

        echo("{} starting up...".format(instance))

        # Check up on its status every so often
        status = instance.update()
        while status == 'pending':
            time.sleep(10)
            status = instance.update()

        if status == 'running':
            for k, v in tags.items():
                instance.add_tag(k, v)
            echo("{} running at {}".format(instance, instance.private_ip_address))
        else:
            secho("Instance was not created correctly", fg="red")
            sys.exit(1)


def _copy_image_command(args):
    _copy_image(args.ami)


def _copy_image(ami_id):

    conf = get_drift_config()
    domain = conf.domain.get()
    aws_region = domain['aws']['ami_baking_region']

    # Grab the source AMI
    source_ami = boto3.resource('ec2', region_name=aws_region).Image(ami_id)

    # Create a list of all regions that are active
    active_tiers = conf.table_store.get_table('tiers').find({'state': 'active'})
    regions = set([tier['aws']['region'] for tier in active_tiers if 'aws' in tier])
    if aws_region in regions:
        regions.remove(aws_region)  # This is the source region
    echo("Distributing {} to region(s) {}.".format(source_ami.id, ', '.join(regions)))

    jobs = []
    for region_id in regions:
        ec2_client = boto3.client('ec2', region_name=region_id)

        ret = ec2_client.copy_image(
            SourceRegion=aws_region,
            SourceImageId=source_ami.id,
            Name=source_ami.name or "",
            Description=source_ami.description or "",
        )

        job = {
            'id': ret['ImageId'],
            'region_id': region_id,
            'client': ec2_client,
        }

        jobs.append(job)

    # Wait on jobs and copy tags
    for job in jobs:
        ami = boto3.resource('ec2', region_name=job['region_id']).Image(job['id'])
        echo("Waiting on {}...".format(ami.id))
        ami.wait_until_exists(Filters=[{'Name': 'state', 'Values': ['available']}])

        if ami.state != 'available':
            continue
        echo("AMI {id} in {region_id} is available. Copying tags...".format(**job))
        job['client'].create_tags(Resources=[job['id']], Tags=source_ami.tags)

    secho("All done.", fg="green")
